import traceback
from dataclasses import dataclass
from typing import Optional, Tuple

import matplotlib.pyplot as plt

HAS_FASTAI = True

# Try to import the necessary modules
# Exception will turn the HAS_FASTAI flag to false
# so that relevant exception can be raised
try:
    import torch
    import numpy as np
    from fastai.basic_train import Learner, RecordOnCPU
    from fastai.vision import Image
    from fastai.torch_core import (
        grab_idx,
        subplots,
        to_detach,
    )
    from fastai.basic_data import DatasetType
    from arcgis.learn.models._unet_utils import show_results_multispectral
    from arcgis.learn._utils.common import dynamic_range_adjustment,kwarg_fill_none
    from arcgis.learn.models._unet_utils import ArcGISSegmentationLabelList
    from arcgis.learn._data_utils._road_orient_data import (
        _plotOrientationOnImage,
        _to_np,
    )
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


@dataclass
class MultiTaskRoadLearner(Learner):
    """show_results
    Custom Multi-Task Learner for Road Orientation training flow.
    """

    def show_results(self, ds_type=DatasetType.Valid, rows: int = 5, **kwargs):
        """
        Show `rows` result of predictions on `ds_type` dataset.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    number of rows of data to be displayed, if
                                batch size is smaller than the rows will
                                display the number provided for batch size.
        =====================   ===========================================
        **Keyword Arguments**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        alpha                   Opacity parameter for label overlay on image
                                float [0..1]
        ---------------------   -------------------------------------------
        """

        total_len = len(self.dl(ds_type))
        batch_size = self.dl(ds_type).batch_size
        self.return_fig = kwargs.get("return_fig", False)
        rows = min(total_len * batch_size, rows)
        data_iterator = iter(self.dl(ds_type))
        ds = self.data.dl(DatasetType.Valid).dataset.orig_data.dl(DatasetType.Valid).dataset
        norm = getattr(self.data.dl(DatasetType.Valid).dataset.orig_data, 'norm', False)

        #for index in range(1, rows+1, batch_size):
            #n_items = (rows + 1 - index) if (index + batch_size) > (rows + 1) else batch_size
            #self.callbacks.append(RecordOnCPU())
            #batch = next(data_iterator)
            #preds = self.pred_batch(ds_type, batch=batch)
            #*self.callbacks, rec_cpu = self.callbacks
            #x, labels = rec_cpu.input, rec_cpu.target
            #x = to_detach(x)
            #labels = [to_detach(y) for y in labels]
            #if norm:
            #    x = self.data.dl(DatasetType.Valid).dataset.orig_data.denorm(x)
            #    if norm.keywords.get('do_y', False):
            #        labels = self.data.dl(DatasetType.Valid).dataset.orig_data.denorm(labels, do_x=True)
            #        preds = self.data.dl(DatasetType.Valid).dataset.orig_data.denorm(preds, do_x=True)
            #xs = [Image(x[i]) for i in range(n_items)]
            #ys = [ds.x.reconstruct(grab_idx(labels[0].unsqueeze(1), i)) for i in range(n_items)]
            #zs = [ds.x.reconstruct(grab_idx(torch.max(preds[0],axis=0)[1].unsqueeze(1),i)) for i in range(n_items)]
            #self._show_xyzs(xs, ys, zs, **kwargs)
            #ds.x.show_xyzs(xs, ys, zs, **kwargs)
            #show_results_multispectral(4,alpha=0.1)
            #self._show_pairs(xs, ys, zs, bin_size=self.data.orient_bin_size, **kwargs)
        total_len = len(self.dl(ds_type))
        batch_size = self.dl(ds_type).batch_size
        rows = min(total_len * batch_size, rows)
        data_iterator = iter(self.dl(ds_type))
        for index in range(1, rows + 1, batch_size):
            n_items = (rows + 1 - index) if (index + batch_size) > (rows + 1) else batch_size
            self.callbacks.append(RecordOnCPU())
            batch = next(data_iterator)
            preds = self.pred_batch(ds_type, batch=batch)
            *self.callbacks, rec_cpu = self.callbacks
            x, labels = rec_cpu.input, rec_cpu.target

            x = to_detach(x)
            labels = [to_detach(y).squeeze(1) for y in labels]

            xs = [grab_idx(x, i).numpy().transpose(1, 2, 0).astype(np.uint8) for i in range(n_items)]
            ys = []
            for i in range(n_items):
                ys.append([_to_np(grab_idx(y, i)) for y in labels])

            zs = []
            for i in range(n_items):
                sub_zs = []
                for pred in preds:
                    _, pred_y = torch.max(pred, axis=1)
                    pred_y = pred_y.squeeze(1)
                    sub_zs.append(_to_np(grab_idx(pred_y, i)))
                zs.append(sub_zs)

            fig, axs = self._show_pairs(xs, ys, zs, bin_size=self.data.orient_bin_size, **kwargs)
            if self.return_fig:
                return fig

    def _show_pairs(
            self,
            xs,
            ys,
            zs,
            imgsize: int = 4,
            figsize: Optional[Tuple[int, int]] = None,
            bin_size: int = 10,
            **kwargs
    ):
        self.alpha = kwargs.get("alpha", 0.6)
        rows = len(xs)
        main_title = "Ground Truth  / Predictions"
        #axs = subplots(rows, 2, imgsize=imgsize, figsize=figsize, title=main_title)
        fig, axs = plt.subplots(nrows=rows, ncols=2, figsize=(2 * imgsize, rows * imgsize))
        fig.suptitle(main_title)
        for x, y, z, ax in zip(xs, ys, zs, axs):

            ax[0].imshow(x)
            ax[0].imshow(y[0], alpha=self.alpha,cmap='binary')
            # _plotOrientationOnImage(ax[0], y[0], x, bin_size)

            ax[1].imshow(x)
            ax[1].imshow(z[0], alpha=self.alpha,cmap='binary')
            # _plotOrientationOnImage(ax[3], z[1], x, bin_size)
        for ax in axs.flatten():
            ax.axis("off")
        # plt.tight_layout()
        #self.display = kwargs.get("for_display", True)
        return fig,axs





from fastai.vision import ItemBase, ItemList, Tensor, ImageList, Tuple, Path, get_transforms, random, open_image, Image, math, plt, torch, Learner, partial, optim, ifnone
from .._utils.common import ArcGISImageList, ArcGISMSImage
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models
import torch
from scipy import linalg
from torch.nn.functional import adaptive_avg_pool2d
import numpy as np
from .._utils.cyclegan import get_activations, InceptionV3
import os
import json

class ImageTuple(ItemBase):
    def __init__(self, img1, img2):
        self.img1,self.img2 = img1,img2
        self.obj,self.data = (img1,img2),[-1+2*img1.data,-1+2*img2.data]
        self.data2 = [-1+2*img2.data,-1+2*img1.data]
        self.shape = img1.shape
    
    def apply_tfms(self, tfms, **kwargs):
        self.img1 = self.img1.apply_tfms(tfms, **kwargs)
        self.img2 = self.img2.apply_tfms(tfms, **kwargs)
        return self
    
    def to_one(self):
        return ArcGISMSImage(0.5+torch.cat(self.data,2)/2)
    def to_one_pred(self): 
        return ArcGISMSImage(0.5+(self.data2[0])/2)
    
    def __repr__(self):
         return f'{self.__class__.__name__}{(self.img1.shape, self.img2.shape)}'

class TargetTupleList(ItemList):
    def reconstruct(self, t:Tensor): 
        if len(t.size()) == 0: return t
        return ImageTuple(ArcGISMSImage(t[0]/2+0.5),ArcGISMSImage(t[1]/2+0.5))

class ImageTupleList2(ImageList):
    _label_cls=TargetTupleList
    def __init__(self, items, itemsB=None, itemsB_valid=None, **kwargs):
        self.itemsB = itemsB
        self.itemsB_valid = itemsB_valid
        super().__init__(items, **kwargs)
    
    def new(self, items, **kwargs):
        return super().new(items, itemsB=self.itemsB, itemsB_valid=self.itemsB_valid, **kwargs)
    
    def get(self, i):
        
        if len(self.items) == len(self.itemsB):
            img1 = super().get(i)
            fn = self.itemsB[i]
        else:
            img1 = super().get(i)
            fn = self.itemsB_valid[i]
        return ImageTuple(img1, open_image(fn))
    
    def reconstruct(self, t:Tensor): 
        return ImageTuple(Image(t[0]/2+0.5),Image(t[1]/2+0.5))
    
    @classmethod
    def from_folders(cls, path, folderA, folderB, **kwargs):
        itemsB = ImageList.from_folder(path/folderB).items
        res = super().from_folder(path/folderA, itemsB=itemsB, itemsB_valid =itemsB, **kwargs)
        res.path = path
        return res
    
    def split_by_idxs(self, train_idx, valid_idx):
        "Split the data between `train_idx` and `valid_idx`."
        self.itemsB_valid = self.itemsB_valid[valid_idx]
        self.itemsB = self.itemsB[train_idx]
        return self.split_by_list(self[train_idx], self[valid_idx])
    
    def show_xys(self, xs, ys, figsize:Tuple[int,int]=(12,6), **kwargs):
        "Show the `xs` and `ys` on a figure of `figsize`. `kwargs` are passed to the show method."
        rows = int(math.sqrt(len(xs)))
        fig, axs = plt.subplots(rows,rows,figsize=figsize)
        for i, ax in enumerate(axs.flatten() if rows > 1 else [axs]):
            xs[i].to_one().show(ax=ax, **kwargs)
        plt.tight_layout()
    
    def show_xyzs(self, xs, ys, zs, figsize:Tuple[int,int]=None, **kwargs):
        """Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`.
        `kwargs` are passed to the show method."""
        
        figsize = ifnone(figsize, (12,3*len(xs)))
        fig,axs = plt.subplots(len(xs), 2, figsize=figsize)
        fig.suptitle('Ground truth / Predictions', weight='bold', size=14)
        for i,(x,z) in enumerate(zip(xs,zs)):
            x.to_one().show(ax=axs[i,0], **kwargs)
            z.to_one_pred().show(ax=axs[i,1], **kwargs)
    
_batch_stats_a = None
_batch_stats_b = None

class ImageTupleListMS2(ArcGISImageList):
    _label_cls=TargetTupleList
    def __init__(self, items, itemsB=None, itemsB_valid=None, **kwargs):
        self.itemsB = itemsB
        self.itemsB_valid = itemsB_valid
        super().__init__(items, **kwargs)

    def new(self, items, **kwargs):
        return super().new(items, itemsB=self.itemsB, itemsB_valid=self.itemsB_valid, **kwargs)

    def get(self, i):
        
        if len(self.items) == len(self.itemsB):
            fn1 = self.items[i]
            img1 = ArcGISMSImage.open(fn1)
            fn = self.itemsB[i]
            img2 = ArcGISMSImage.open(fn)
            if img1.shape[0] < img2.shape[0]:
                cont = []
                last_tile = np.expand_dims(img1.data[img1.shape[0]-1,:,:], 0)
                res = abs(img2.shape[0] - img1.shape[0])
                for i in range(res):
                    img1 = ArcGISMSImage(torch.tensor(np.concatenate((img1.data, last_tile), axis=0)))
            if img2.shape[0] < img1.shape[0]:
                cont = []
                last_tile = np.expand_dims(img2.data[img2.shape[0]-1,:,:], 0)
                res = abs(img1.shape[0] - img2.shape[0])
                for i in range(res):
                    img2 = ArcGISMSImage(torch.tensor(np.concatenate((img2.data, last_tile), axis=0)))
        else:
            fn1 = self.items[i]
            img1 = ArcGISMSImage.open(fn1)
            fn = self.itemsB_valid[i]
            img2 = ArcGISMSImage.open(fn)
            if img1.shape[0] < img2.shape[0]:
                cont = []
                last_tile = np.expand_dims(img1.data[img1.shape[0]-1,:,:], 0)
                res = abs(img2.shape[0] - img1.shape[0])
                for i in range(res):
                    img1 = ArcGISMSImage(torch.tensor(np.concatenate((img1.data, last_tile), axis=0)))
            if img2.shape[0] < img1.shape[0]:
                cont = []
                last_tile = np.expand_dims(img2.data[img2.shape[0]-1,:,:], 0)
                res = abs(img1.shape[0] - img2.shape[0])
                for i in range(res):
                    img2 = ArcGISMSImage(torch.tensor(np.concatenate((img2.data, last_tile), axis=0)))
        global _batch_stats_a
        global _batch_stats_b
        img1_scaled = _tensor_scaler_tfm(img1.data, min_values=_batch_stats_a['band_min_values'], max_values=_batch_stats_a['band_max_values'], mode='minmax')
        img2_scaled = _tensor_scaler_tfm(img2.data, min_values=_batch_stats_b['band_min_values'], max_values=_batch_stats_b['band_max_values'], mode='minmax')
        img1_scaled = ArcGISMSImage(img1_scaled)
        img2_scaled = ArcGISMSImage(img2_scaled)
        return ImageTuple(img1_scaled, img2_scaled)
    
    def reconstruct(self, t:Tensor): 
        return ImageTuple(ArcGISMSImage(t[0]/2+0.5),ArcGISMSImage(t[1]/2+0.5))
    
    @classmethod
    def from_folders(cls, path, folderA, folderB, batch_stats_a, batch_stats_b, **kwargs):
        itemsB = ImageList.from_folder(path/folderB).items
        res = super().from_folder(path/folderA, itemsB=itemsB, itemsB_valid=itemsB, **kwargs)
        res.path = path
        global _batch_stats_a
        global _batch_stats_b
        _batch_stats_a = batch_stats_a
        _batch_stats_b = batch_stats_b
        return res
    
    def split_by_idxs(self, train_idx, valid_idx):
        "Split the data between `train_idx` and `valid_idx`."
        self.itemsB_valid = self.itemsB_valid[valid_idx]
        self.itemsB = self.itemsB[train_idx]
        return self.split_by_list(self[train_idx], self[valid_idx])
    
    def show_xys(self, xs, ys, figsize:Tuple[int,int]=(12,6), **kwargs):
        "Show the `xs` and `ys` on a figure of `figsize`. `kwargs` are passed to the show method."
        rows = int(math.sqrt(len(xs)))
        fig, axs = plt.subplots(rows,rows,figsize=figsize)
        for i, ax in enumerate(axs.flatten() if rows > 1 else [axs]):
            xs[i].to_one().show(ax=ax, **kwargs)
        plt.tight_layout()

    def show_xyzs(self, xs, ys, zs, figsize:Tuple[int,int]=None, **kwargs):
        """Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`.
        `kwargs` are passed to the show method."""
        
        figsize = ifnone(figsize, (12,3*len(xs)))
        fig,axs = plt.subplots(len(xs), 2, figsize=figsize)
        fig.suptitle('Ground truth / Predictions', weight='bold', size=14)
        for i,(x,z) in enumerate(zip(xs,zs)):
            x.to_one().show(ax=axs[i,0], **kwargs)
            z.to_one_pred().show(ax=axs[i,1], **kwargs)

def calculate_activation_statistics(batch_size, data_len, batch_list):
    act = get_activations(batch_size, data_len, batch_list)
    mu = np.mean(act, axis=0)
    sigma = np.cov(act, rowvar=False)
    return mu, sigma

def _tensor_scaler_tfm(tensor_batch, min_values, max_values, mode='minmax'):
    from .._data import _tensor_scaler
    x = tensor_batch
    if x.shape[0] > min_values.shape[0]:
        res = x.shape[0] - min_values.shape[0]
        last_val = torch.tensor([min_values[min_values.shape[0]-1]])
        for i in range(res):
            min_values = torch.tensor(np.concatenate((min_values, last_val), axis=0))
    if x.shape[0] > max_values.shape[0]:
        res = x.shape[0] - max_values.shape[0]
        last_val = torch.tensor([max_values[max_values.shape[0]-1]])
        for i in range(res):
            max_values = torch.tensor(np.concatenate((max_values, last_val), axis=0))
    max_values = max_values.view(-1, 1, 1).to(x.device)
    min_values = min_values.view(-1, 1, 1).to(x.device)
    x = _tensor_scaler(x, min_values, max_values, mode, create_view=False)
    return x

def _batch_stats_json(path, img_list, norm_pct, stats_file_name="esri_normalization_stats.json"):
    from .._data import _get_batch_stats
    if len(img_list) < 300:
        norm_pct = 1

    dummy_stats = {
                "batch_stats_for_norm_pct_0" : {
                    "band_min_values":None, 
                    "band_max_values":None, 
                    "band_mean_values":None, 
                    "band_std_values":None, 
                    "scaled_min_values":None, 
                    "scaled_max_values":None, 
                    "scaled_mean_values":None, 
                    "scaled_std_values":None}}

    normstats_json_path = os.path.abspath(path / '..' / stats_file_name)

    if not os.path.exists(normstats_json_path):       
        normstats = dummy_stats
        with open(normstats_json_path, 'w', encoding='utf-8') as f:
            json.dump(normstats, f, ensure_ascii=False, indent=4)
    else:
        with open(normstats_json_path) as f:
                normstats = json.load(f)

    norm_pct_search = f"batch_stats_for_norm_pct_{round(norm_pct*100)}"
    if norm_pct_search in normstats:
        batch_stats = normstats[norm_pct_search]
        for s in batch_stats:
            if batch_stats[s] is not None:
                batch_stats[s] = torch.tensor(batch_stats[s])
    else:
        batch_stats = _get_batch_stats(img_list, norm_pct)
        normstats[norm_pct_search] = dict(batch_stats)
        for s in normstats[norm_pct_search]:
            if normstats[norm_pct_search][s] is not None:
                normstats[norm_pct_search][s] = normstats[norm_pct_search][s].tolist()
        with open(normstats_json_path, 'w', encoding='utf-8') as f:
            json.dump(normstats, f, ensure_ascii=False, indent=4)

    return batch_stats

def prepare_data_ms_pix2pix(path, norm_pct, val_split_pct, seed, databunch_kwargs):
    folder_a = 'train_a'
    folder_b = 'train_b'

    img_list_a = ArcGISImageList.from_folder(path/"train_a")
    img_list_b = ArcGISImageList.from_folder(path/"train_b")
    
    batch_stats_a = _batch_stats_json(path, img_list_a, norm_pct, stats_file_name="esri_normalization_stats_a.json")
    batch_stats_b = _batch_stats_json(path, img_list_b, norm_pct, stats_file_name="esri_normalization_stats_b.json")

    data = ImageTupleListMS2.from_folders(path, folder_a, folder_b, batch_stats_a, batch_stats_b)\
            .split_by_rand_pct(val_split_pct, seed=seed)\
            .label_empty()\
            .databunch(**databunch_kwargs)

    data._band_min_values = batch_stats_a['band_min_values']
    data._band_max_values = batch_stats_a['band_max_values']
    data._band_mean_values = batch_stats_a['band_mean_values']
    data._band_std_values = batch_stats_a['band_std_values']
    data._scaled_min_values = batch_stats_a['scaled_min_values']
    data._scaled_max_values = batch_stats_a['scaled_max_values']
    data._scaled_mean_values = batch_stats_a['scaled_mean_values']
    data._scaled_std_values = batch_stats_a['scaled_std_values']

    data._band_min_values_b = batch_stats_b['band_min_values']
    data._band_max_values_b = batch_stats_b['band_max_values']
    data._band_mean_values_b = batch_stats_b['band_mean_values']
    data._band_std_values_b = batch_stats_b['band_std_values']
    data._scaled_min_values_b = batch_stats_b['scaled_min_values']
    data._scaled_max_values_b = batch_stats_b['scaled_max_values']
    data._scaled_mean_values_b = batch_stats_b['scaled_mean_values']
    data._scaled_std_values_b = batch_stats_b['scaled_std_values']

    return data


import os
import math
import torch
import matplotlib.pyplot as plt
import matplotlib
from ..models._maskrcnn_utils import ArcGISImageSegment
from .common import get_nbatches, kwarg_fill_none, image_batch_stretcher

def show_batch_rcnn_masks(self, rows=3, alpha=0.5, **kwargs): # parameters adjusted in kwargs
    nrows = rows
    ncols = kwarg_fill_none(kwargs, 'ncols', 3)
    
    n_items = None
    if kwargs.get('n_items', None) is not None:
        n_items = kwargs.get('n_items')

    type_data_loader = kwargs.get('data_loader', 'training') # options : traininig, validation, testing
    if type_data_loader == 'training':
        data_loader = self.train_dl
    elif type_data_loader == 'validation':
        data_loader = self.valid_dl
    elif type_data_loader == 'testing':
        data_loader = self.test_dl
    else:
        e = Exception(f'could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing ')
        raise(e)

    imsize = 5
    if kwargs.get('imsize', None) is not None:
        imsize = kwargs.get('imsize')

    cmap = kwargs.get('cmap', 'tab20')

    statistics_type = kwargs.get('statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get('stretch_type', 'minmax') # Accepted Values `minmax`, `percentclip`

    # Get Batch
    if n_items is None:
        n_items = nrows*ncols
    else:
        nrows = math.ceil(n_items/ncols)
    n_items = min(n_items, len(self.x))

    x_batch, y_batch = get_nbatches(data_loader, n_items)
    x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)

    if self._is_multispectral:
        # Denormalize X
        if self._do_normalize:
            x_batch = (self._scaled_std_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch) * x_batch ) + self._scaled_mean_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch)

        rgb_bands = self._symbology_rgb_bands
        if kwargs.get('rgb_bands', None) is not None:
            rgb_bands = kwargs.get('rgb_bands')

        nodata = 0
        if kwargs.get('nodata', None) is not None:
            nodata = kwargs.get('nodata')

        e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
        symbology_bands = []
        if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
            raise(e)
        for b in rgb_bands:
            if type(b) == str:
                b_index = self._bands.index(b)
            elif type(b) == int:
                self._bands[b] # To check if the band index specified by the user really exists.
                b_index = b
            else:
                raise(e)
            b_index = self._extract_bands.index(b_index)
            symbology_bands.append(b_index)

        # Extract RGB Bands
        symbology_x_batch = x_batch[:, symbology_bands]
        if stretch_type is not None:
            symbology_x_batch = image_batch_stretcher(symbology_x_batch, stretch_type, statistics_type)

        # Channel first to channel last and clamp float values to range 0 - 1 for plotting
        symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
        # Clamp float values to range 0 - 1
        if symbology_x_batch.mean() < 1:
            symbology_x_batch = symbology_x_batch.clamp(0, 1)

        # Get color Array
        color_array = self._multispectral_color_array
        color_array[1:, 3] = alpha
    else:
        symbology_x_batch = x_batch.permute(0, 2, 3, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Size for plotting
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            axi = axs
            if nrows == 1:
                axi = axi
            else:
                axi = axi[r]
            if ncols == 1:
                axi = axi
            else:
                axi = axi[c]
            axi.axis('off')
            if idx < symbology_x_batch.shape[0]:
                axi.imshow(symbology_x_batch[idx].cpu().numpy())
                n_instance = y_batch[idx].unique().shape[0]
                y_merged = y_batch[idx].max(dim=0)[0].cpu().numpy()
                cmap_fn = getattr(matplotlib.cm, cmap)
                y_rgba = cmap_fn._resample(n_instance)(y_merged)
                y_rgba[y_merged == 0] = 0
                y_rgba[:, :, -1] = alpha
                axi.imshow(y_rgba)
                axi.axis('off')
            idx+=1


import torch
import matplotlib.pyplot as plt
import math
from .common import get_nbatches, image_batch_stretcher

def show_batch_labeled_tiles(self, rows=3, **kwargs): # parameters adjusted in kwargs   
    nrows = rows
    ncols = kwargs.get('ncols', nrows)
    #start_index = kwargs.get('start_index', 0) # Does not work with dataloader
    
    n_items = kwargs.get('n_items', nrows*ncols)
    n_items = min(n_items, len(self.x))
    nrows = math.ceil(n_items/ncols)
    nbatches = math.ceil(n_items/self.batch_size)

    type_data_loader = kwargs.get('data_loader', 'training') # options : traininig, validation, testing
    if type_data_loader == 'training':
        data_loader = self.train_dl
    elif type_data_loader == 'validation':
        data_loader = self.valid_dl
    elif type_data_loader == 'testing':
        data_loader = self.test_dl
    else:
        e = Exception(f'could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing ')
        raise(e)

    rgb_bands = kwargs.get('rgb_bands', self._symbology_rgb_bands)
    nodata = kwargs.get('nodata', 0)
    imsize = kwargs.get('imsize', 5)
    statistics_type = kwargs.get('statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get('stretch_type', 'minmax') # Accepted Values `minmax`, `percentclip`

    e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
    symbology_bands = []
    if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
        raise(e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = self._bands.index(b)
        elif type(b) == int:
            self._bands[b] # To check if the band index specified by the user really exists.
            b_index = b
        else:
            raise(e)
        b_index = self._extract_bands.index(b_index)
        symbology_bands.append(b_index)

    # Get Batch
    x_batch, y_batch = get_nbatches(data_loader, nbatches)
    x_batch = torch.cat(x_batch)
    # Denormalize X
    x_batch = (self._scaled_std_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch) * x_batch ) + self._scaled_mean_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch)
    y_batch = torch.cat(y_batch)

    # Extract RGB Bands
    symbology_x_batch = x_batch[:, symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(symbology_x_batch, stretch_type, statistics_type)

    # Channel first to channel last and clamp float values to range 0 - 1 for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._multispectral_color_array

    # Size for plotting
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            if idx < symbology_x_batch.shape[0]:
                axi = axs
                if nrows == 1:
                    axi = axi
                else:
                    axi = axi[r]
                if ncols == 1:
                    axi = axi
                else:
                    axi = axi[c]
                axi.imshow(symbology_x_batch[idx].cpu().numpy())

                if self.dataset_type == "MultiLabeled_Tiles":
                    one_hot_labels = y_batch[idx].tolist()
                    from itertools import compress
                    labels = compress(self.classes, one_hot_labels)
                    title = ";".join(labels)
                else:
                    title = f"{self.classes[y_batch[idx].item()]}"
                
                axi.set_title(title)
                axi.axis('off')
            else:
                ax[r][c].axis('off')
            idx += 1



from ._wms import WMSLayer
from .wmts import WMTSLayer
from ._csv import CSVLayer
from ._georss import GeoRSSLayer
from ._kml import KMLLayer
from ._geojson import GeoJSONLayer

__all__ = ['WMTSLayer', 'CSVLayer', 'GeoRSSLayer',
           'KMLLayer', 'WMSLayer', 'GeoJSONLayer']


import os
import sys
import traceback


HAS_BACKEND_SET = False
ARCGIS_ENABLE_TF_BACKEND = os.environ.get('ARCGIS_ENABLE_TF_BACKEND') is '1'
LAMBDA_TEXT_CLASSIFICATION = os.environ.get('LAMBDA_TEXT_CLASSIFICATION') is '1'

HAS_TENSORFLOW = False
tf_import_exception = None



class FakeImport():
    def __getattr__(self, attr):
        return self
    def __call__(self, *args, **kwargs):
        return self

if LAMBDA_TEXT_CLASSIFICATION:
    default_module = FakeImport()
    missing_modules = ['scipy','scipy.stats', 'spacy','spacy.symbols','spacy.blank', 'matplotlib',\
        'matplotlib.pyplot', 'matplotlib.patches', 'matplotlib.cm', 'scipy.special', 'PIL']
    for module_name in missing_modules: 
        sys.modules[module_name] = default_module

try:
    if ARCGIS_ENABLE_TF_BACKEND:
        import tensorflow as tf
        HAS_TENSORFLOW = True
except Exception as e:
    tf_import_exception = traceback.format_exc()
    pass

def enable_backend():
    global HAS_BACKEND_SET
    global HAS_TENSORFLOW
    global ARCGIS_ENABLE_TF_BACKEND

    if ARCGIS_ENABLE_TF_BACKEND:
        #if tf.__version__ == '2.0.0':
        if HAS_TENSORFLOW  and not HAS_BACKEND_SET:
            #tf.keras.backend.set_image_data_format('channels_first')
            try:
                tf.compat.v1.enable_eager_execution()
            except:
                pass
            tf_set_gpu_memory_growth()
            tf_sample_op()
            tf.keras.backend.clear_session()
            HAS_BACKEND_SET = True

def tf_set_gpu_memory_growth():
    gpus = tf.config.experimental.list_physical_devices('GPU')
    if gpus:
        try:
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices('GPU')
            #print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            print(e)

def raise_tensorflow_import_error():
    if ARCGIS_ENABLE_TF_BACKEND:
        message = """
        Could not find tensorflow, Please install tensorflow using the following command
        \nconda install -c esri tensorflow-gpu=2.1.0
        """
        ex = Exception(message)
        raise(ex)
    else:
        message = """
        Please enable tensorflow by setting the required environment variable 'ARCGIS_ENABLE_TF_BACKEND' to '1' before importing arcgis
        \n for example the following code block needs to be executed before importing arcgis
        \n\n`import os; os.environ['ARCGIS_ENABLE_TF_BACKEND'] = '1'`
        """
        ex = Exception(message)
        raise(ex)


def tf_sample_op():
    a = tf.keras.layers.Conv2D(1, (3, 3))
    a = a(tf.zeros((1, 3, 20, 20))).numpy()


## Fastai Imports #######

HAS_FASTAI = False
fastai_import_exception = None

def do_fastai_imports():
    global HAS_FASTAI
    global fastai_import_exception

    try:
        import fastai
        import torch
        import torchvision
        import skimage
        HAS_FASTAI = True
    except Exception as e:
        fastai_import_exception = traceback.format_exc()
        pass

    try:
        from .patches import precondition
    except:
        pass


def fastai_installation_command():
    installation_steps = "Install them using 'conda install -c esri arcgis=1.8.1 pillow scikit-image'\n'conda install -c fastai -c pytorch fastai pytorch=1.4.0 torchvision=0.5.0 tensorflow-gpu=2.1.0'\n'conda install gdal=2.3.3'"

    return installation_steps

def raise_fastai_import_error(import_exception=fastai_import_exception, installation_steps=None, message=None):
    if installation_steps is None:
        installation_steps = fastai_installation_command()
    if message is None:
        message = "This module requires fastai, PyTorch, torchvision and scikit-image as its dependencies."
    raise Exception(f"""{import_exception} \n\n{message}\n{installation_steps}""")


HAS_GDAL = False
gdal_import_exception = None
GDAL_INSTALL_MESSAGE = f"""
\nPlease install gdal using the following command
\nconda install gdal=2.3.3
""".strip()

try:
    from osgeo import gdal
    HAS_GDAL = True
except Exception as e:
    gdal_import_exception = traceback.format_exc()
    pass

def raise_gdal_import_error(import_exception=gdal_import_exception):
    message = "gdal is required to work with multispectral datasets."
    raise Exception(f"""{import_exception} \n\n{message}\n{GDAL_INSTALL_MESSAGE}""")



import os
import tempfile
import traceback
import json
import warnings
import math
from pathlib import Path

HAS_FASTAI = True
import_exception=None

import arcgis
from arcgis.features import FeatureLayer

try:
    from ._arcgis_model import ArcGISModel, _raise_fastai_import_error
    from fastai.tabular import tabular_learner, TabularDataBunch, TabularList, TabularModel
    from fastai.tabular.transform import FillMissing, Categorify, Normalize
    from fastai.basic_train import Learner, load_learner
    from fastprogress.fastprogress import progress_bar
    from .._utils.tabular_data import TabularDataObject
    from .._utils.common import _get_emd_path
    from fastai.torch_core import split_model_idx
    import torch
    from fastai.metrics import r2_score
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False

HAS_NUMPY = True
try:
    import numpy as np
except:
    HAS_NUMPY = False


def _get_learner_object(data, layers, emb_szs, ps, emb_drop, pretrained_path):

    if pretrained_path:
        learn = load_learner(os.path.dirname(pretrained_path),
                             os.path.basename(pretrained_path).split('.')[0] + "_exported.pth")
        if not data._is_empty:
            learn.data = data._databunch
    else:
        databunch = data._databunch
        if not emb_szs or isinstance(emb_szs, dict):
            emb_szs = databunch.get_emb_szs({} if not emb_szs else emb_szs)

        model = TabularModel(emb_szs, len(databunch.cont_names), out_sz=databunch.c, layers=layers, ps=ps, emb_drop=emb_drop,
                             y_range=None, use_bn=False)
        learn = Learner(databunch, model, model_dir=tempfile.TemporaryDirectory().name)

    return learn


class FullyConnectedNetwork(ArcGISModel):
    """
    Creates a FullyConnectedNetwork Object.
    Based on the Fast.ai's Tabular Learner

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            `prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    layers                  Optional list, specifying the number of nodes in each layer.
                            Default: [500, 100] is used.
                            2 layers each with nodes 500 and 100 respectively.
    ---------------------   -------------------------------------------
    emb_szs                 Optional dict, variable name with embedding size
                            for categorical variables.
                            If not specified, then calculated using fastai.
    =====================   ===========================================

    :returns: `FullyConnectedNetwork` Object
    """

    def __init__(self, data, layers=None, emb_szs=None, **kwargs):

        if data._is_unsupervised:
            raise Exception("Cannot train on unsupervised data")

        super().__init__(data, None)

        self._backbone = None
        if layers is None:
            layers = [500, 100]

        ps = kwargs.get('ps', None)
        emb_drop = kwargs.get('emb_drop', 0)

        self.learn = _get_learner_object(data, layers, emb_szs, ps, emb_drop, kwargs.get('pretrained_path', None))
        self._layers = layers
        self.learn.model = self.learn.model.to(self._device)
        idx = 1
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=1e-03)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a FullyConnectedNetwork Object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_tabulardata` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `FullyConnectedNetwork` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        dependent_variable = emd['dependent_variable']
        categorical_variables = emd['categorical_variables']
        continuous_variables = emd['continuous_variables']
        _is_classification = False
        if emd['_is_classification'] == "classification":
            _is_classification = True

        layers = emd['layers']
        if data is None:

            data = TabularDataObject._empty(categorical_variables, continuous_variables, dependent_variable, None)
            data._is_classification = _is_classification
            class_object = cls(data, pretrained_path=emd_path)
            class_object._data.emd = emd
            class_object._data.emd_path = emd_path
            return class_object

        return cls(data, layers=layers, pretrained_path=str(emd_path))

    def save(self, name_or_path, framework='PyTorch', publish=False, gis=None, save_optimizer=False, **kwargs):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by ``SingleShotDetector``, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional GIS Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        """

        if '\\' in name_or_path or '/' in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(os.getcwd(), name_or_path)

        if not os.path.exists(path):
            os.mkdir(path)

        self.learn.export(os.path.join(path, os.path.basename(path) + '_exported.pth'))
        from IPython.utils import io
        with io.capture_output() as captured:
            super().save(path, framework, publish, gis, save_optimizer=save_optimizer, **kwargs)

        return Path(path)

    @property
    def _model_metrics(self):
        from IPython.utils import io
        with io.capture_output() as captured:
            score = self.score()

        return {'score': score}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["ModelType"] = "FullyConnectedNetwork"
        _emd_template["layers"] = self._layers
        _emd_template["dependent_variable"] = self._data._dependent_variable
        _emd_template["categorical_variables"] = self._data._categorical_variables
        _emd_template["continuous_variables"] = self._data._continuous_variables
        _emd_template['_is_classification'] = "classification" if self._data._is_classification else "regression"

        return _emd_template

    def _predict(self, data_frame_row):
        return self.learn.predict(data_frame_row)

    def _df_predict(self, dataframe):
        # from fastai.data_block import split_kwargs_by_func, grab_idx, DatasetType
        # ds_type = DatasetType.Valid
        # ds = self.learn.dl(ds_type).dataset
        #
        # current_databunch = self.learn.data
        # databunch_half, databunch_second_half = self._data._prepare_validation_databunch(dataframe)
        # return dataframe, databunch_half, databunch_second_half
        # try:
        #     self.learn.data = databunch
        #     preds = self.learn.get_preds(ds_type)[0]
        # except Exception as e:
        #     raise e
        # finally:
        #     self.learn.data = current_databunch
        #
        # analyze_kwargs, kwargs = split_kwargs_by_func({}, ds.y.analyze_pred)
        # preds = [ds.y.analyze_pred(grab_idx(preds, i), **analyze_kwargs) for i in range(len(preds))]
        # preds = [ds.y.reconstruct(z) for z in preds]
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        preds = []

        for i in progress_bar(range(len(dataframe))):
            prediction = self._predict(dataframe.iloc[i])[0].obj
            if isinstance(prediction, (list, np.ndarray)):
                prediction = prediction[0]
            preds.append(prediction)

        return preds

    def predict(
            self,
            input_features=None,
            explanatory_rasters=None,
            datefield=None,
            distance_features=None,
            output_layer_name="Prediction Layer",
            gis=None,
            prediction_type='features',
            output_raster_path=None,
            match_field_names=None):
        """

        Predict on data from feature layer, dataframe and or raster data.

        =================================   =========================================================================
        **Argument**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional Feature Layer or spatially enabled dataframe.
                                            Required if prediction_type='features'.
                                            Contains features with location and
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list of Raster Objects.
                                            If prediction_type='raster', must contain all rasters
                                            required to make predictions.
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional string. Field name from feature layer
                                            that contains the date, time for the input features.
                                            Same as `prepare_tabulardata()`.
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of Feature Layer objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as `prepare_tabulardata()`.
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional GIS Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output feature layer predictions.
                                            With this feature_layer argument is required.

                                            Set 'raster', to make prediction raster.
                                            With this rasters must be specified.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path.
                                            Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional dictionary.
                                            Specify mapping of field names from prediction set
                                            to training set.
                                            For example:
                                                {
                                                    "Field_Name_1": "Field_1",
                                                    "Field_Name_2": "Field_2"
                                                }
        =================================   =========================================================================

        :returns Feature Layer if prediction_type='features', dataframe for prediction_type='dataframe' else creates an output raster.

        """

        rasters = explanatory_rasters if explanatory_rasters else []
        if prediction_type in ['features', 'dataframe']:

            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(input_features, rasters, datefield, distance_features, output_layer_name, gis, match_field_names, prediction_type)
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception("Please specify output_raster_folder_path to save the output.")

            return self._predict_rasters(output_raster_path, rasters, match_field_names)

    def _predict_features(
            self,
            input_features,
            rasters=None,
            datefield=None,
            distance_feature_layers=None,
            output_name="Prediction Layer",
            gis=None,
            match_field_names=None,
            prediction_type="features"
    ):
        if isinstance(input_features, FeatureLayer):
            dataframe = input_features.query().sdf
        else:
            dataframe = input_features.copy()

        fields_needed = self._data._categorical_variables + self._data._continuous_variables
        distance_feature_layers = distance_feature_layers if distance_feature_layers else []

        continuous_variables = self._data._continuous_variables

        feature_layer_columns = []
        for column in dataframe.columns:
            column_name = column
            categorical = False

            if column_name in fields_needed:
                if column_name not in continuous_variables:
                    categorical = True
            elif match_field_names and match_field_names.get(column_name):
                if match_field_names.get(column_name) not in continuous_variables:
                    categorical = True
            else:
                continue

            feature_layer_columns.append((column_name, categorical))

        raster_columns = []
        if rasters:
            for raster in rasters:
                column_name = raster.name
                categorical = False
                if column_name in fields_needed:
                    if column_name not in continuous_variables:
                        categorical = True
                elif match_field_names and match_field_names.get(column_name):
                    column_name = match_field_names.get(column_name)
                    if column_name not in continuous_variables:
                        categorical = True
                else:
                    continue

                raster_columns.append((raster, categorical))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            processed_dataframe, fields_mapping = TabularDataObject._prepare_dataframe_from_features(
                input_features,
                self._data._dependent_variable,
                feature_layer_columns,
                raster_columns,
                datefield,
                distance_feature_layers
            )

        if match_field_names:
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed:
            if field not in processed_dataframe.columns:
                raise Exception(f"Field missing {field}")

        dataframe["prediction_results"] = self._df_predict(processed_dataframe.copy())
        if prediction_type == "dataframe":
            return dataframe

        return dataframe.spatial.to_featurelayer(output_name, gis)

    def _predict_rasters(self, output_folder_path, rasters, match_field_names=None):

        if not os.path.exists(os.path.dirname(output_folder_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_folder_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        fields_needed = self._data._continuous_variables + self._data._categorical_variables

        try:
            arcpy.env.outputCoordinateSystem = rasters[0].extent['spatialReference']['wkt']
        except:
            arcpy.env.outputCoordinateSystem = rasters[0].extent['spatialReference']['wkid']

        xmin = rasters[0].extent['xmin']
        xmax = rasters[0].extent['xmax']
        ymin = rasters[0].extent['ymin']
        ymax = rasters[0].extent['ymax']
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = rasters[0].extent['spatialReference']

        for raster in rasters:
            point_upper = arcgis.geometry.Point(
                {'x': raster.extent['xmin'], 'y': raster.extent['ymax'], 'sr': raster.extent['spatialReference']})
            point_lower = arcgis.geometry.Point(
                {'x': raster.extent['xmax'], 'y': raster.extent['ymin'], 'sr': raster.extent['spatialReference']})
            cell_size = arcgis.geometry.Point(
                {'x': raster.mean_cell_width, 'y': raster.mean_cell_height, 'sr': raster.extent['spatialReference']})

            points = arcgis.geometry.project([point_upper, point_lower, cell_size], raster.extent['spatialReference'],
                                             default_sr)
            point_upper = points[0]
            point_lower = points[1]
            cell_size = points[2]

            if xmin > point_upper.x:
                xmin = point_upper.x
            if ymax < point_upper.y:
                ymax = point_upper.y
            if xmax < point_lower.x:
                xmax = point_lower.x
            if ymin > point_lower.y:
                ymin = point_lower.y

            if min_cell_size_x > cell_size.x:
                min_cell_size_x = cell_size.x

            if min_cell_size_y > cell_size.y:
                min_cell_size_y = cell_size.y

        max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))

        point_upper = arcgis.geometry.Point({'x': xmin, 'y': ymax, 'sr': default_sr})
        cell_size = arcgis.geometry.Point({'x': min_cell_size_x, 'y': min_cell_size_y, 'sr': default_sr})

        raster_data = {}
        for raster in rasters:
            field_name = raster.name
            point_upper_translated = arcgis.geometry.project([point_upper], default_sr, raster.extent['spatialReference'])[0]
            cell_size_translated = arcgis.geometry.project([cell_size], default_sr, raster.extent['spatialReference'])[0]
            if field_name in fields_needed:
                raster_read = raster.read(origin_coordinate=(point_upper_translated.x, point_upper_translated.y), ncols=max_raster_columns, nrows=max_raster_rows, cell_size=(cell_size_translated.x, cell_size_translated.y))
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            elif match_field_names and match_field_names.get(raster.name):
                field_name = match_field_names.get(raster.name)
                raster_read = raster.read(origin_coordinate=(point_upper_translated.x, point_upper_translated.y), ncols=max_raster_columns, nrows=max_raster_rows, cell_size=(cell_size_translated.x, cell_size_translated.y))
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            else:
                continue

        for field in fields_needed:
            if field not in list(raster_data.keys()) and match_field_names and match_field_names.get(field, None) is None:
                raise Exception(f"Field missing {field}")

        processed_data = []

        length_values = len(raster_data[list(raster_data.keys())[0]])
        for i in range(length_values):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_data.append(processed_row)

        processed_numpy = np.array(self._df_predict(pd.DataFrame(data=np.array(processed_data), columns=sorted(raster_data))), dtype='float64')
        processed_numpy = processed_numpy.reshape([max_raster_rows, max_raster_columns])
        processed_raster = arcpy.NumPyArrayToRaster(processed_numpy, arcpy.Point(xmin, ymin), x_cell_size=min_cell_size_x, y_cell_size=min_cell_size_y)
        processed_raster.save(output_folder_path)

        return True

    def show_results(self, rows=5):
        """
        Prints the rows of the dataframe with target and prediction columns.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================

        :returns: dataframe
        """
        self._check_requisites()

        validation_dataframe = self._data._dataframe.loc[self._data._validation_indexes].reset_index(drop=True)

        if rows > len(validation_dataframe):
            rows = len(validation_dataframe)

        rows_df = validation_dataframe.loc[:rows].copy()

        predictions = self._df_predict(rows_df)
        rows_df['prediction_results'] = predictions

        return rows_df

    def score(self):
        """
        :returns R2 score for regression model and Accuracy for classification model.
        """

        self._check_requisites()
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        validation_dataframe = self._data._dataframe.loc[self._data._validation_indexes].reset_index(drop=True)

        predictions = np.array(self._df_predict(validation_dataframe))
        labels = validation_dataframe[self._data._dependent_variable]

        if self._data._is_classification:
            return (np.array(predictions)==labels).mean()
        else:
            return float(r2_score(torch.tensor(np.array(predictions, dtype='float64')), torch.tensor(np.array(labels, dtype='float64'))))



import torch
import numpy as np
import math
import torch.nn.functional as F


def calculate_intersection(preds, targs, mode): 
    """
    Calculates intersection between preds and targs.
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor. 
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor. 
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean' 
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns intersection between preds and targs -> torch.tensor
    """

    if mode == 'per_class': 
        dim = 2
        assert targs.ndim == 3
    if mode == 'mean': 
        dim = 1
        assert targs.ndim == 2

    return (preds * targs).sum(dim=dim)


def calculate_union(preds, targs, intersection, mode): 
    """
    Calculates union between preds and targs.
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor. 
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor. 
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    intersection            Required torch.tensor. 
                            Intersection between preds and targs
                            calculate_intersection(preds, targs, mode)
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean' 
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns union between preds and targs -> torch.tensor
    """
    if mode == 'per_class': 
        dim = 2
        assert targs.ndim == 3
    if mode == 'mean': 
        dim = 1
        assert targs.ndim == 2

    return (preds + targs).sum(dim=dim) - intersection

def confusion_matrix(preds, targs, intersection, mode):
    """
    Calculates true positive, false positive and false negetive
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor. 
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor. 
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean' 
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns true positive, false positive and false negetive -> torch.tensor, torch.tensor, torch.tensor
    """

    if mode == 'per_class': 
        dim = 2
        assert targs.ndim == 3
    if mode == 'mean': 
        dim = 1
        assert targs.ndim == 2

    TP = intersection                                #true positive
    FP = preds.sum(dim=dim) - intersection           #flase positive
    FN = targs.sum(dim=dim) - intersection           #false negetive

    return TP, FP, FN

def calculate_precision(TP, FP, eps=1e-08):
    return TP / (TP + FP + eps)

def calculate_recall(TP, FN, eps=1e-08):
    return TP / (TP + FN + eps)

def calculate_f1(precision, recall, eps=1e-08):
    return 2 * (precision * recall) / (precision + recall + eps)

def prepare_output_for_metrics(preds, targs, ignore_classes, mode):
    """
    This function modifies the shape of targs(ground truth segmentation mask) 
    to match with preds(predicted output) 
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor. 
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor. 
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean' 
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns modeified preds and targs 
    """

    from arcgis.learn._utils.segmentation_loss_functions import  expand_outputs
    num = preds.size(0)
    classes = preds.size(1)
    keep_classes = [v for v in range(classes) if v not in ignore_classes]
    encoded_targs = expand_outputs(preds, targs)
    preds = preds.argmax(dim=1)
    preds = F.one_hot(preds, classes).permute(0, 3, 1, 2)
    if mode == 'per_class':
        preds = preds.contiguous().float().view(num, classes, -1)[:, keep_classes, :]
        encoded_targs = encoded_targs.float().view(num, classes, -1)[:, keep_classes, :]
    if mode == 'mean':
        preds = preds.contiguous().float().view(num, classes, -1)[:, keep_classes, :].view(num,-1)
        encoded_targs = encoded_targs.float().view(num, classes, -1)[:, keep_classes, :].view(num,-1)
    return  preds, encoded_targs

def calculate_metrics(preds, targs, mode,  ignore_classes=[]):
    """
    Calculates precision, recall and f1 scores.
    """
    # import pdb; pdb.set_trace();
    preds, targs = prepare_output_for_metrics(preds, targs, ignore_classes, mode)
    intersection = calculate_intersection(preds, targs, mode)
    union = calculate_union(preds, targs, intersection, mode)
    TP, FP, FN = confusion_matrix(preds, targs, intersection, mode) 
    precision = calculate_precision(TP, FP, eps=1e-08)
    recall = calculate_recall(TP, FN, eps=1e-08)
    f1 = calculate_f1(precision, recall, eps=1e-08)
    return precision, recall, f1


def post_process_CD(preds, y, n_classes):
    from .change_detection_data import post_process, post_process_y
    preds = post_process(preds)
    y[y == 1] = 0
    y[y == -1] = 1
    preds = preds.squeeze(1)
    preds = F.one_hot(preds, n_classes).permute(0, 3, 1, 2)
    return preds, y


def per_class_metrics(self,
                      ignore_classes=[], 
                      **kwargs):
    """
    Computes per class precision, recall and f1 scores 
    """
    import pandas as pd
    dl = kwargs.get('dl', None)
    postprocess_type = kwargs.get('postprocess_type', None)
    ignore_mapped_class = kwargs.get('ignore_mapped_class', ignore_classes)
    keep_classes = [c for i,c in  enumerate(self._data.classes) if i not in ignore_mapped_class]
    model = self.learn.model.eval()
    precision_list = []
    recall_list = []
    f1_list = []
    for batch in self._data.valid_dl if dl is None else dl:
        x, y = batch
        y = y.to('cpu')
        with torch.no_grad():
            if type(x) is list or type(x) is tuple:
                predictions = model(*x).detach().to('cpu')
            else:
                predictions = model(x).detach().to('cpu')
            if postprocess_type == 'CD':
                predictions, y = post_process_CD(predictions, 
                                                 y, 
                                                 n_classes=len(self._data.classes))

            precision, recall, f1 = calculate_metrics(predictions, 
                                                      y, 
                                                      mode='per_class', 
                                                      ignore_classes=ignore_mapped_class)
            precision_list.append(precision.mean(0))
            recall_list.append(recall.mean(0))
            f1_list.append(f1.mean(0))

    precision=torch.stack(tuple(precision_list)).mean(0)  
    recall=torch.stack(tuple(recall_list)).mean(0)   
    f1=torch.stack(tuple(f1_list)).mean(0)     

    columns = keep_classes
    df= pd.DataFrame(columns=columns)
    df.loc['precision'] = precision.tolist()
    df.loc['recall'] = recall.tolist()
    df.loc['f1'] = f1.tolist()
    return df


def show_batch_classified_tiles(self, rows=3, alpha=0.7, **kwargs):
    import matplotlib.pyplot as plt
    from .._utils.common import kwarg_fill_none, get_nbatches, find_data_loader, get_top_padding, get_symbology_bands, dynamic_range_adjustment, denorm_x, image_tensor_checks_plotting

    imsize = kwarg_fill_none(kwargs, 'imsize', 5)
    nrows = rows
    ncols = kwarg_fill_none(kwargs, 'ncols', 3)
    imsize = kwarg_fill_none(kwargs, 'imsize', 5)
    statistics_type = kwarg_fill_none(kwargs, 'statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
        
    n_items = kwargs.get('n_items', nrows*ncols)
    n_items = min(n_items, len(self.x))
    nrows = math.ceil(n_items/ncols)

    top = kwargs.get('top', None)
    title_font_size=16
    if top is None:
        top = get_top_padding(
            title_font_size=title_font_size, 
            nrows=nrows, 
            imsize=imsize
            )

    # Get n batches
    type_data_loader = kwarg_fill_none(kwargs, 'data_loader', 'training') # options : traininig, validation, testing
    data_loader = find_data_loader(type_data_loader, self)
    x_batch, y_batch = get_nbatches(data_loader, math.ceil(n_items/self.batch_size))
    symbology_x_batch = x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)


    symbology_bands = [0, 1, 2]
    if self._is_multispectral:
        # Get RGB Bands for plotting
        rgb_bands = kwarg_fill_none(kwargs, 'rgb_bands', self._symbology_rgb_bands)

        # Get Symbology bands
        symbology_bands = get_symbology_bands(rgb_bands, self._extract_bands, self._bands)

    # Denormalize X
    x_batch = denorm_x(x_batch, self)

    # Extract RGB Bands for plotting
    symbology_x_batch = x_batch[:, symbology_bands]

    # Apply Image Strecthing
    if statistics_type == 'DRA':
        symbology_x_batch = dynamic_range_adjustment(symbology_x_batch)

    symbology_x_batch = image_tensor_checks_plotting(symbology_x_batch)

    # Get color Array
    color_array = self._multispectral_color_array
    color_array[1:, 3] = alpha

    # Plot now
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            axi = axs
            if nrows == 1:
                axi = axi
            else:
                axi = axi[r]
            if ncols == 1:
                axi = axi
            else:
                axi = axi[c]
            axi.axis('off')
            if idx < symbology_x_batch.shape[0]:
                axi.imshow(symbology_x_batch[idx].cpu().numpy())
                y_rgb = color_array[y_batch[idx][0]].cpu().numpy()
                axi.imshow(y_rgb, alpha=alpha)
            idx+=1




from pathlib import Path

from arcgis.learn._data import _bb_pad_collate, prepare_data
from arcgis.learn._data_utils._base_data import ArcgisData
from arcgis.learn._data_utils._pixel_classifier_data import ClassifiedTilesData

__data_classes__ = {
    "Classified_Tiles": ClassifiedTilesData
}


class ArcgisDataFactory:
    def __init__(
        self, path, class_mapping=None, val_split_pct=0.1, batch_size=64,
        transforms=None, collate_fn=_bb_pad_collate, seed=42, **kwargs
    ):
        self.path = path
        self.class_mapping = class_mapping
        self.val_split_pct = val_split_pct
        self.batch_size = batch_size
        self.transforms = transforms
        self.collate_fn = collate_fn
        self.seed = seed
        self.kwargs = kwargs

    def create(self, dataset_type):
        if dataset_type in __data_classes__.keys():
            return __data_classes__[dataset_type](
                path=self.path, class_mapping=self.class_mapping, val_split_pct=self.val_split_pct,
                batch_size=self.batch_size, transforms=self.transforms, seed=self.seed,
                dataset_type=dataset_type, **self.kwargs,
            )
        else:
            return None


def prepare_data_future(
    path, class_mapping=None, chip_size=224, val_split_pct=0.1, batch_size=64,
    transforms=None, collate_fn=_bb_pad_collate, seed=42, dataset_type=None,
    resize_to=None, **kwargs
):
    """
    Prepares a data object from training sample exported by the
    Export Training Data tool in ArcGIS Pro or Image Server, or training
    samples in the supported dataset formats. This data object consists of
    training and validation data sets with the specified transformations,
    chip size, batch size, split percentage, etc.
    -For object detection, use Pascal_VOC_rectangles or KITTI_rectangles format.
    -For feature categorization use Labelled Tiles or ImageNet format.
    -For pixel classification, use Classified Tiles format.
    -For entity extraction from text, use IOB, BILUO or ner_json formats.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    path                    Required string. Path to data directory.
    ---------------------   -------------------------------------------
    class_mapping           Optional dictionary. Mapping from id to
                            its string label.
                            For dataset_type=IOB, BILUO or ner_json:
                                Provide address field as class mapping
                                in below format:
                                class_mapping={'address_tag':'address_field'}.
                                Field defined as 'address_tag' will be treated
                                as a location. In cases where trained model extracts
                                multiple locations from a single document, that
                                document will be replicated for each location.

    ---------------------   -------------------------------------------
    chip_size               Optional integer, default 224. Size of the image to train the
                            model. Images are cropped to the specified chip_size. If image size is less
                            than chip_size, the image size is used as chip_size.
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size for mini batch gradient
                            descent (Reduce it if getting CUDA Out of Memory
                            Errors).
    ---------------------   -------------------------------------------
    transforms              Optional tuple. Fast.ai transforms for data
                            augmentation of training and validation datasets
                            respectively (We have set good defaults which work
                            for satellite imagery well). If transforms is set
                            to `False` no transformation will take place and
                            `chip_size` parameter will also not take effect.
                            If the dataset_type is 'PointCloud', use
                            `Transform3d` class from `arcgis.learn`.
    ---------------------   -------------------------------------------
    collate_fn              Optional function. Passed to PyTorch to collate data
                            into batches(usually default works).
    ---------------------   -------------------------------------------
    seed                    Optional integer. Random seed for reproducible
                            train-validation split.
    ---------------------   -------------------------------------------
    dataset_type            Optional string. `prepare_data` function will infer
                            the `dataset_type` on its own if it contains a
                            map.txt file. If the path does not contain the
                            map.txt file pass either of 'PASCAL_VOC_rectangles',
                            'KITTI_rectangles', 'RCNN_Masks', 'Classified_Tiles',
                            'Labeled_Tiles', 'Imagenet', 'PointCloud' and
                            'ImageCaptioning'.
    ---------------------   -------------------------------------------
    resize_to               Optional integer. Resize the image to given size.
    =====================   ===========================================

    **Keyword Arguments**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    imagery_type            Optional string. Type of imagery used to export
                            the training data, valid values are:
                                - 'naip'
                                - 'sentinel2'
                                - 'landsat8'
                                - 'ms' - any other type of imagery
    ---------------------   -------------------------------------------
    bands                   Optional list. Bands of the imagery used to export
                            training data.
                            For example ['r', 'g', 'b', 'nir', 'u']
                            where 'nir' is near infrared band and 'u' is a miscellaneous band.
    ---------------------   -------------------------------------------
    rgb_bands               Optional list. Indices of red, green and blue bands
                            in the imagery used to export the training data.
                            for example: [2, 1, 0]
    ---------------------   -------------------------------------------
    extract_bands           Optional list. Indices of bands to be used for
                            training the model, same as in the imagery used to
                            export the training data.
                            for example: [3, 1, 0] where we will not be using
                            the band at index 2 to train our model.
    ---------------------   -------------------------------------------
    norm_pct                Optional float. Percentage of training data to be
                            used for calculating imagery statistics for
                            normalizing the data.
                            Default is 0.3 (30%) of data.
    ---------------------   -------------------------------------------
    downsample_factor       Optional integer. Factor to downsample the images
                            for image SuperResolution.
                            for example: if value is 2 and image size 256x256,
                            it will create label images of size 128x128.
                            Default is 4
    ---------------------   -------------------------------------------
    sub_dataset_type        Optional string. Special sub dataset type required
                            for creating the Orientation data to use for
                            training Multi-Task Road Extractor. Currently support
                            Classified Tiles dataset format.
                            Value : "RoadOrientation"
    ---------------------   -------------------------------------------
    road_extractor_params   Optional dict. Parameters to use in Multi-Task
                                            road extractor.
                            - gaussian_thresh: set the gaussian threshold which
                                                allows to set the required road width
                            - orient_bin_size: set the bin size for orientation angles
                            - orient_theta: set the width of orientation mask
                            - is_gaussian_mask: flag to set, if given labels are converted
                                                to gaussian. if false, system will create the
                                                gaussian mask internally and convert to binary
                                                mask of appropriate road width based on
                                                the gaussian_thresh
    =====================   ===========================================

    :returns: data object
    """
    has_esri_files = ArcgisData._check_esri_files(Path(path))
    if has_esri_files and dataset_type is None:
        dataset_type = ArcgisData._get_dataset_type(Path(path))
    if has_esri_files and dataset_type in __data_classes__.keys():
        dataset_type = (
            dataset_type
            if dataset_type
            else ArcgisData._get_dataset_type(Path(path))
        )
        objData = __data_classes__[dataset_type](
            path=path, class_mapping=class_mapping, chip_size=chip_size, val_split_pct=val_split_pct,
            batch_size=batch_size, transforms=transforms, seed=seed,
            dataset_type=dataset_type, resize_to=resize_to, **kwargs,
        )
        return objData.get_databunch(**kwargs)
    else:
        return prepare_data(
            path, class_mapping, chip_size, val_split_pct, batch_size, transforms,
            collate_fn, seed, dataset_type, resize_to, **kwargs
        )



# Referred from https://github.com/tcapelle/timeseries_fastai
import torch
import torch.nn as nn
from .layers import *


class Shortcut(nn.Module):
    def __init__(self, ni, nf, device=None):
        super().__init__()
        if ni == nf:
            self.conv = noop
        else:
            self.conv = convlayer(ni, nf, ks=1, act_fn=False)

    def forward(self, x):
        return self.conv(x)


class _TSResNet(nn.Module):
    def __init__(self, input, output, device=None):
        super().__init__()
        conv_sizes = [64, 128, 128]
        kss = [7, 5, 3]

        self.layers = []
        self.block1 = nn.ModuleList([
            convlayer(input, conv_sizes[0], kss[0], act_fn='relu'),
            convlayer(conv_sizes[0], conv_sizes[0], kss[1], act_fn='relu'),
            convlayer(conv_sizes[0], conv_sizes[0], kss[2], act_fn=False),
            Shortcut(input, conv_sizes[0]),
            nn.ReLU()
        ])

        self.block2 = nn.ModuleList([
            convlayer(conv_sizes[0], conv_sizes[1], kss[0], act_fn='relu'),
            convlayer(conv_sizes[1], conv_sizes[1], kss[1], act_fn='relu'),
            convlayer(conv_sizes[1], conv_sizes[1], kss[2], act_fn=False),
            Shortcut(conv_sizes[0], conv_sizes[1]),
            nn.ReLU()
        ])

        self.block3 = nn.ModuleList([
            convlayer(conv_sizes[1], conv_sizes[2], kss[0], act_fn='relu'),
            convlayer(conv_sizes[2], conv_sizes[2], kss[1], act_fn='relu'),
            convlayer(conv_sizes[2], conv_sizes[2], kss[2], act_fn=False),
            Shortcut(conv_sizes[1], conv_sizes[2]),
            nn.ReLU()
        ])

        self.blocks = [self.block1, self.block2, self.block3]

        self.avg_pooling = nn.AdaptiveAvgPool1d(1)
        self.linear = nn.Linear(conv_sizes[-1], output)

    def forward(self, x):
        for block in self.blocks:
            orig = x
            for layer in block:
                if not isinstance(layer, Shortcut):
                    x = layer(x)
                else:
                    x = x + layer(orig)

        x = self.avg_pooling(x).squeeze(-1)
        output = self.linear(x)

        return output



try:
    import fastai
    from fastai.torch_core import *
    from fastai.basic_data import *
    from fastai.data_block import *
    from fastai.core import *
    import torch
except:
    raise Exception("Unable to import")

import arcgis

if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
    device = torch.device("cuda")
elif getattr(arcgis.env, "_processorType", "") == "CPU":
    device = torch.device("cpu")
else:
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')


class TSItem(ItemBase):
    "`ItemBase` suitable for time series"

    def __init__(self, item, *args, **kwargs):
        super().__init__(item, *args, **kwargs)
        self.data = item
        self.obj = item
        self.channels = item.shape[-2]
        self.seq_len = item.shape[-1]


    def __str__(self):
        return 'TimeSeries(ch={:.0f}, seq_len={:.0f})'.format(
            self.channels, self.seq_len)

    def clone(self):
        return self.__class__(self.data.clone())

    def apply_tfms(self, tfms=None, **kwargs):
        if tfms is None: return self
        x = self.clone()
        for tfm in tfms: x.data = tfm(x.data)
        return x

    def reconstruct(self, item):
        return TSItem(item)

    def show(self, ax=None, title=None, **kwargs):
        x = self.clone()
        if ax is None:
            plt.plot(x.data.transpose_(0, 1))
            plt.title(title)
            plt.show()
        else:
            ax.plot(x.data.transpose_(0, 1))
            ax.title.set_text(title)
            ax.tick_params(
                axis='both',
                which='both',
                bottom='off',
                top='off',
                labelbottom='off',
                right='off',
                left='off',
                labelleft='off')
            return ax


class TimeSeriesItem(TSItem): pass


class TSDataBunch(DataBunch):

    def scale(self, scale_type='standardize', scale_by_channel=False, scale_by_sample=False,
          scale_range=(-1, 1)) -> None:
        self.scale_type = scale_type
        self.scale_by_channel = scale_by_channel
        self.scale_by_sample = scale_by_sample
        self.scale_range = scale_range
        if scale_type is None:
            self.stats = None
            return self
        assert scale_type in ['normalize', 'standardize', 'robustscale'], print('Select a correct type', scale_type)

        train = self.train_ds.x.items.astype(float)
        valid = self.valid_ds.x.items.astype(float)
        if self.test_ds is not None: test = self.test_ds.x.items.astype(float)

        if scale_by_channel and scale_by_sample: axis = -1 # mean
        elif scale_by_channel: axis = (0, 2) # mean for the entire dataset by channel
        elif scale_by_sample: axis = (1, 2)  # mean for each sample
        else: axis = None


        if scale_by_sample:
            self.stats = None
            if scale_type == 'normalize':
                train_min = np.nanmin(train, axis=axis, keepdims=True)
                train_max = np.nanmax(train, axis=axis, keepdims=True)
                self.train_ds.x.items = (((train - train_min)) / (train_max - train_min)) *\
                (self.scale_range[1] - self.scale_range[0]) + self.scale_range[0]
                valid_min = np.nanmin(valid, axis=axis, keepdims=True)
                valid_max = np.nanmax(valid, axis=axis, keepdims=True)
                self.valid_ds.x.items = (((valid - valid_min)) / (valid_max - valid_min)) *\
                (self.scale_range[1] - self.scale_range[0]) + self.scale_range[0]
                if self.test_ds is not None:
                    test_min = np.nanmin(test, axis=axis, keepdims=True)
                    test_max = np.nanmax(test, axis=axis, keepdims=True)
                    self.test_ds.x.items = (((test - test_min)) / (test_max - test_min)) *\
                    (self.scale_range[1] - self.scale_range[0]) + self.scale_range[0]
                return self

            elif scale_type == 'standardize':
                train_mean = np.nanmean(train, axis=axis, keepdims=True)
                train_std = np.nanstd(train, axis=axis, keepdims=True) + 1e-8
                self.train_ds.x.items = (train - train_mean) / train_std
                valid_mean = np.nanmean(valid, axis=axis, keepdims=True)
                valid_std = np.nanstd(valid, axis=axis, keepdims=True) + 1e-8
                self.valid_ds.x.items = (valid - valid_mean) / valid_std
                if self.test_ds is not None:
                    test_mean = np.nanmean(test, axis=axis, keepdims=True)
                    test_std = np.nanstd(test, axis=axis, keepdims=True) + 1e-8
                    self.test_ds.x.items = (test - test_mean) / test_std
                return self

            elif scale_type == 'robustscale':
                train_median = np.nanmedian(train, axis=axis, keepdims=True)
                train_perc_25 = np.nanpercentile(train, 25, axis=axis, keepdims=True)
                train_perc_75 = np.nanpercentile(train, 75, axis=axis, keepdims=True)
                train_scale = train_perc_75 - train_perc_25
                self.train_ds.x.items = (train - train_median) / train_scale

                valid_median = np.nanmedian(valid, axis=axis, keepdims=True)
                valid_perc_25 = np.nanpercentile(valid, 25, axis=axis, keepdims=True)
                valid_perc_75 = np.nanpercentile(valid, 75, axis=axis, keepdims=True)
                valid_scale = valid_perc_75 - valid_perc_25
                self.valid_ds.x.items = (valid - valid_median) / valid_scale

                if self.test_ds is not None:
                    test_median = np.nanmedian(test, axis=axis, keepdims=True)
                    test_perc_25 = np.nanpercentile(test, 25, axis=axis, keepdims=True)
                    test_perc_75 = np.nanpercentile(test, 75, axis=axis, keepdims=True)
                    test_scale = test_perc_75 - test_perc_25
                    self.test_ds.x.items = (test - test_median) / test_scale
                return self

        else:
            if scale_type == 'normalize':
                train_min = np.nanmin(train, axis=axis, keepdims=True)
                train_max = np.nanmax(train, axis=axis, keepdims=True)
                self.stats = train_min, train_max
                self.train_ds.x.items = (((self.train_ds.x.items - train_min)) / (train_max - train_min)) *\
                (self.scale_range[1] - self.scale_range[0]) + self.scale_range[0]
                self.valid_ds.x.items = (((self.valid_ds.x.items - train_min)) /  (train_max - train_min)) *\
                (self.scale_range[1] - self.scale_range[0]) + self.scale_range[0]
                if self.test_ds is not None:
                    self.test_ds.x.items = (((self.test_ds.x.items - train_min)) / (train_max - train_min)) *\
                    (self.scale_range[1] - self.scale_range[0]) + self.scale_range[0]
                return self
            elif scale_type == 'standardize':
                train_mean = np.nanmean(train, axis=axis, keepdims=True)
                train_std = np.nanstd(train, axis=axis, keepdims=True) + 1e-8
                self.stats = train_mean, train_std
                self.train_ds.x.items = (self.train_ds.x.items - train_mean) / train_std
                self.valid_ds.x.items = (self.valid_ds.x.items - train_mean) / train_std
                if self.test_ds is not None:
                    self.test_ds.x.items = (self.test_ds.x.items - train_mean) / train_std
                return self
            elif scale_type == 'robustscale':
                train_median = np.nanmedian(train, axis=axis, keepdims=True)
                train_perc_25 = np.nanpercentile(train, 25, axis=axis, keepdims=True)
                train_perc_75 = np.nanpercentile(train, 75, axis=axis, keepdims=True)
                train_scale = train_perc_75 - train_perc_25
                self.stats = train_median, train_scale
                self.train_ds.x.items = (train - train_median) / train_scale
                self.valid_ds.x.items = (valid - train_median) / train_scale
                if self.test_ds is not None:  self.test_ds.x.items = (test - train_median) / train_scale
                return self

    @property
    def cw(self)->None: return self._get_cw(self.train_dl)

    @property
    def dbtype(self)->str: return '1D'

    def _get_cw(self, train_dl):
        target = torch.Tensor(train_dl.dataset.y.items).to(dtype=torch.int64)
        # Compute samples weight (each sample should get its own weight)
        class_sample_count = torch.tensor(
            [(target == t).sum() for t in torch.unique(target, sorted=True)])
        weights = 1. / class_sample_count.float()
        return (weights / weights.sum()).to(device)


def show_counts(databunch):
    labels, counts = np.unique(databunch.train_ds.y.items, return_counts=True)
    plt.bar(labels, counts)
    plt.title('labels')
    plt.xticks(labels)
    plt.show()


DataBunch.show_counts = show_counts


class TSPreProcessor(PreProcessor):

    def __init__(self, ds: ItemList): self.ds = ds

    def process(self, ds: ItemList):
        ds.features, ds.seq_len = self.ds.get(0).data.size(-2), self.ds.get(0).data.size(-1)
        ds.f = ds.features
        ds.s = ds.seq_len


class TimeSeriesList(ItemList):
    "`ItemList` suitable for time series"
    _bunch = TSDataBunch
    _processor = TSPreProcessor
    _label_cls = None
    _square_show = True

    def __init__(self, items, *args, mask=None, tfms=None, **kwargs):
        items = To3dTensor(items)
        super().__init__(items, *args, **kwargs)
        self.tfms,self.mask = tfms,mask
        self.copy_new.append('tfms')

    def get(self, i):
        item = super().get(i)
        if self.mask is None: return TSItem(To2dTensor(item))
        else: return[TSItem(To2dTensor(item[m])) for m in self.mask]


    def show_xys(self, xs, ys, figsize=(10, 10), **kwargs):
        "Show the `xs` and `ys` on a figure of `figsize`. `kwargs` are passed to the show method."
        rows = int(math.sqrt(len(xs)))
        fig, axs = plt.subplots(rows, rows, figsize=figsize)
        for x, y, ax in zip(xs, ys, axs.flatten()):
            x.show(ax=ax, title=str(y), **kwargs)
        plt.tight_layout()
        plt.show()

    def show_xyzs(self, xs, ys, zs, figsize=(10, 10), **kwargs):
        if self._square_show_res:
            rows = int(np.ceil(math.sqrt(len(xs))))
            fig, axs = plt.subplots(
                rows,
                rows,
                figsize=figsize
            )
            fig.suptitle('Ground truth vs Predictions', fontsize=16)
            for x, y, z, ax in zip(xs, ys, zs, axs.flatten()):
                x.show(ax=ax, title=f'{str(y)}\n{str(z)}', **kwargs)
        else:
            fig, axs = plt.subplots(
                len(xs),
                2,
                figsize=figsize
            )
            fig.suptitle('Ground truth vs Predictions', fontsize=16)
            for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
                x.show(ax=axs[i, 0], title=str(y), **kwargs)
                x.show(ax=axs[i, 1], title=str(z), **kwargs)
        plt.tight_layout()
        plt.show()

    @classmethod
    def from_array(cls, ts, **kwargs):
        return cls(ts)

    @classmethod
    def from_df(cls, df, path='.', cols=None, feat=None, processor=None, **kwargs) -> 'ItemList':
        "Create an `ItemList` in `path` from the inputs in the `cols` of `df`."
        if cols is None:
            inputs = df
        else:
            cols = listify(cols)
            if feat is not None and feat not in cols: cols = cols + listify(feat)
            col_idxs = df_names_to_idx(list(cols), df)
            inputs = df.iloc[:, col_idxs]
        assert inputs.isna().sum().sum(
        ) == 0, f"You have NaN values in column(s) {cols} of your dataframe, please fix it."
        inputs = df2array(inputs, feat)
        res = cls(
            items=inputs,
            path=path,
            inner_df=df,
            processor=processor,
            **kwargs)
        return res


class TSList(TimeSeriesList): pass


def df2array(df, feat=None):
    if feat is None:
        return df.values[:, None]
    for i, ch in enumerate(df[feat].unique()):
        data_i = df[df[feat] == ch].values[:, None]
        if i == 0: data = data_i
        else: data = np.concatenate((data, data_i), axis=1)

    return data


def ToTensor(arr, **kwargs):
    if isinstance(arr, np.ndarray):
        arr = torch.from_numpy(arr)
    elif not isinstance(arr, torch.Tensor):
        print(f"Can't convert {type(arr)} to torch.Tensor")
    return arr.float()


def ToArray(arr):
    if isinstance(arr, torch.Tensor):
        arr = np.array(arr)
    elif not isinstance(arr, np.ndarray):
        print(f"Can't convert {type(arr)} to np.array")
    if arr.dtype == 'O': arr = np.array(arr, dtype=np.float32)
    return arr


def To3dTensor(arr):
    if isinstance(arr, list):
        arr = np.array(arr)
    if arr.dtype == 'O': arr = np.array(arr, dtype=np.float32)
    arr = ToTensor(arr)
    if arr.ndim == 1: arr = arr[None, None]
    elif arr.ndim == 2: arr = arr[:, None]
    elif arr.ndim == 4: arr = arr[0]
    assert arr.ndim == 3, 'Please, review input dimensions'
    return arr


def To2dTensor(arr):
    if arr.dtype == 'O': arr = np.array(arr, dtype=np.float32)
    arr = ToTensor(arr)
    if arr.ndim == 1: arr = arr[None]
    elif arr.ndim == 3: arr = torch.squeeze(arr, 0)
    assert arr.ndim == 2, 'Please, review input dimensions'
    return arr


def To1dTensor(arr):
    if arr.dtype == 'O': arr = np.array(arr, dtype=np.float32)
    arr = ToTensor(arr)
    if arr.ndim == 3: arr = torch.squeeze(arr, 1)
    if arr.ndim == 2: arr = torch.squeeze(arr, 0)
    assert arr.ndim == 1, 'Please, review input dimensions'
    return arr


def To3dArray(arr):
    arr = ToArray(arr)
    if arr.ndim == 1: arr = arr[None, None]
    elif arr.ndim == 2: arr = arr[:, None]
    elif arr.ndim == 4: arr = arr[0]
    assert arr.ndim == 3, 'Please, review input dimensions'
    return np.array(arr)


def To2dArray(arr):
    arr = ToArray(arr)
    if arr.ndim == 1: arr = arr[None]
    if arr.ndim == 3: arr = np.squeeze(arr, 0)
    assert arr.ndim == 2, 'Please, review input dimensions'
    return np.array(arr)


def To1dArray(arr):
    arr = ToArray(arr)
    if arr.ndim == 3: arr = np.squeeze(arr, 1)
    if arr.ndim == 2: arr = np.squeeze(arr, 0)
    assert arr.ndim == 1, 'Please, review input dimensions'
    return np.array(arr)


import re
import os
import sys
import copy
import types
import shutil
import random
import logging
import warnings
import traceback
import pandas as pd
from functools import partial


HAS_FASTAI = True
try:
    import torch
    import arcgis
    from fastai.text import TextList, TextClasDataBunch, pad_collate, TextDataBunch,\
            SortishSampler, SortSampler, ItemList, ItemBase, Text
    from fastai.data_block import CategoryList, MultiCategoryList
    from ._seq2seq_utils import SequenceToSequenceTextList, teacher_forcing_tfm
    from .text_transforms import TransformerNERDataset, TransformerNERDataBunch, process_text
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False

HAS_BEAUTIFULSOUP = True
try:
    from bs4 import BeautifulSoup
except:
    HAS_BEAUTIFULSOUP = False
else:
    warnings.filterwarnings("ignore", category=UserWarning, module='bs4')

HAS_NUMPY = True
try:
    import numpy as np
    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
except:
    HAS_NUMPY = False


max_len=100


def _raise_fastai_exception(exception):
    error_message = (f"{exception}\n\nThis module requires fastai, PyTorch and transformers as its dependencies.\n"
                     "Install them using - 'conda install -c esri -c fastai -c pytorch arcgis=1.8.2 "
                          "scikit-image=0.15.0 pillow=6.2.2 libtiff=4.0.10 fastai=1.0.60 pytorch=1.4.0 "
                          "torchvision=0.5.0 scikit-learn=0.23.1 --no-pin'"
                          "\n'conda install gdal=2.3.3'"
                          "\n'pip install transformers==3.3.0'")
    raise Exception(error_message)


# Overriding show_text_xys function of TextList to display the dataframe in desired fashion
def show_text_xys(self, xs, ys, max_len:int=max_len):
    "Show the `xs` (inputs) and `ys` (targets). `max_len` is the maximum number of tokens displayed."
    if not HAS_NUMPY:
        raise Exception("This module requires numpy.")

    from IPython.display import display
    names = ['idx','text'] if self._is_lm else ['text','target']
    items = []
    for i, (x,y) in enumerate(zip(xs,ys)):
        txt_x = ' '.join(x.text.split(' ')[:max_len]) if max_len is not None else x.text
        items.append([i, txt_x] if self._is_lm else [txt_x, y])
    items = np.array(items)
    df = pd.DataFrame({n:items[:,i] for i,n in enumerate(names)}, columns=names)
    dataframe_style = df.style\
        .set_table_styles([dict(selector='th', props=[('text-align', 'left')])])\
        .set_properties(**{'text-align': "left"}).hide_index()
    display(dataframe_style)


# Overriding show_text_xyzs function of TextList to display the dataframe in desired fashion
def show_text_xyzs(self, xs, ys, zs, max_len:int=max_len):
    "Show `xs` (inputs), `ys` (targets) and `zs` (predictions). `max_len` is the maximum number of tokens displayed."
    if not HAS_NUMPY:
        raise Exception("This module requires numpy.")

    from IPython.display import display
    items,names = [],['text','target','prediction']
    for i, (x,y,z) in enumerate(zip(xs,ys,zs)):
        txt_x = ' '.join(x.text.split(' ')[:max_len]) if max_len is not None else x.text
        items.append([txt_x, y, z])
    items = np.array(items)
    df = pd.DataFrame({n:items[:,i] for i,n in enumerate(names)}, columns=names)
    dataframe_style = df.style \
        .set_table_styles([dict(selector='th', props=[('text-align', 'left')])]) \
        .set_properties(**{'text-align': "left"}).hide_index()
    display(dataframe_style)


def read_file(path):
    filename, file_extension = os.path.splitext(path)
    if file_extension == ".csv":
        return pd.read_csv(path, dtype='str')
    else:
        return pd.read_csv(path, sep="\t", dtype='str')


def save_data_in_model_metrics_html(text, path, model_characteristics_folder):
    file_path = os.path.join(path, "model_metrics.html")
    if os.path.exists(file_path):
        with open(file_path, mode="a+", encoding="utf-8") as fp:
            fp.write(text)

    folder_path = os.path.join(path, model_characteristics_folder)
    if os.path.exists(folder_path):
        with open(os.path.join(folder_path, "sample_results.html"), mode="w", encoding="utf-8") as fp:
            fp.write(text)


def copy_metrics(source, target, folder_name):
    source_file_path = os.path.join(source, 'model_metrics.html')
    target_file_path = os.path.join(target, 'model_metrics.html')
    if os.path.exists(source_file_path):
        if os.path.exists(target_file_path): os.remove(target_file_path)
        _ = shutil.copyfile(source_file_path, target_file_path)

    source_folder_path = os.path.join(source, folder_name)
    target_folder_path = os.path.join(target, folder_name)
    if os.path.exists(source_folder_path):
        if os.path.exists(target_folder_path): shutil.rmtree(target_folder_path)
        _ = shutil.copytree(source_folder_path, target_folder_path)


def extract_entities(tokens, labels):
    prev_label, token_list, entities = labels[0], [tokens[0]], []
    for token_index, (token, label) in enumerate(list(zip(tokens[1:], labels[1:]))):
        if label == 'O':
            if prev_label != 'O': entities.append((token_list, prev_label))
            token_list, prev_label = list(), label

        if prev_label == label:
            token_list.append(token)
        else:
            if prev_label != 'O':entities.append((token_list, prev_label))
            token_list = list()
            prev_label = label
            token_list.append(token)
    if token_list:
        entities.append((token_list, prev_label))
    return entities


class TextDataObject:

    def __init__(self, task):
        self.emd = {}
        self.emd_path = None

        self._bs = None
        self._is_empty = True
        self._train_df = None
        self._valid_df = None
        self._text_cols = None
        self._label_cols = list()
        self._databunch = None
        self._training_indexes = list()
        self._task = task
        self._backbone = None
        
        self.databunch_kwargs = {'num_workers': 0} if sys.platform == 'win32' else {}
        self.databunch_kwargs["pin_memory"] = True
        if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
            self.databunch_kwargs["device"] = torch.device("cuda")
        elif getattr(arcgis.env, "_processorType", "") == "CPU":
            self.databunch_kwargs["device"] = torch.device("cpu")
        else:
            self.databunch_kwargs["device"] = torch.device('cuda') if torch.cuda.is_available() else torch.device(
                'cpu')

    @classmethod
    def prepare_data_for_entity_recognition(
            cls,
            tokens_collection,
            tags_collection,
            address_tag,
            unique_tags,
            seed=42,
            batch_size=8,
            val_split_pct = 0.1
        ):

        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        text_data = cls("ner")
        random.seed(seed)
        # temp = list(zip(tokens_collection, tags_collection))
        # random.shuffle(temp)
        # tokens_collection, tags_collection = zip(*temp)

        random.Random(seed).shuffle(tokens_collection)
        random.Random(seed).shuffle(tags_collection)

        # used when creating databunch for ner
        split_index = int((1 - val_split_pct) * len(tokens_collection))

        text_data._bs = batch_size
        text_data._unique_tags = unique_tags
        text_data._address_tag = address_tag
        text_data._label2id = {x: index for index, x in enumerate(unique_tags)}

        text_data._train_tags = tags_collection[:split_index]
        text_data._train_tokens = tokens_collection[:split_index]

        text_data._valid_tags = tags_collection[split_index:]
        text_data._valid_tokens = tokens_collection[split_index:]
        if text_data._address_tag not in text_data._unique_tags:
            logging.warning("No Address tag found in your data.\n\
            1. If your data has an address field, pass your address field name as address tag in class mapping \n\
            e.g. - data=prepare_data(dataset_type=ds_type,path=training_data_folder,\n\t\t\t\
                class_mapping={address_tag:address_field_name})\n\
            2. Else no action is required, if your data does not have any address information.")

        return text_data

    @classmethod
    def prepare_data_for_classification(
            cls,
            data,
            text_cols,
            label_cols,
            train_file="train.csv",
            valid_file=None,
            val_split_pct=0.1,
            seed=42,
            batch_size=8,
            process_labels=False,
            remove_html_tags=False,
            remove_urls=False
        ):
        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        text_data = cls("classification")
        if not os.path.exists(data):
            raise Exception(f"Provided data directory - {data}, does not exists")

        training_file_path = os.path.join(data, train_file)

        if not os.path.exists(training_file_path):
            raise Exception(f"Provided data directory does not contain {train_file} file")

        train_df = read_file(training_file_path)
        train_df = cls._preprocess_df(train_df, text_cols, label_cols, process_labels, remove_html_tags, remove_urls)

        random.seed(seed)

        if valid_file is not None and os.path.exists(os.path.join(data, valid_file)):
            validation_file_exists = True
        else:
            validation_file_exists = False

        if validation_file_exists:
            valid_df = read_file(os.path.join(data, valid_file))
            valid_df = cls._preprocess_df(valid_df, text_cols, label_cols, process_labels, remove_html_tags, remove_urls)
        else:
            validation_indexes = random.sample(range(train_df.shape[0]), round(val_split_pct * train_df.shape[0]))
            training_indexes = list(set([i for i in range(train_df.shape[0])]) - set(validation_indexes))

            temp_df = copy.deepcopy(train_df)
            train_df = temp_df.loc[training_indexes]
            valid_df = temp_df.loc[validation_indexes]
            # Removing rows with empty strings in the text_cols from the training and validation data
            train_df[text_cols].replace('', np.nan, inplace=True)
            train_df.dropna(inplace=True)
            valid_df[text_cols].replace('', np.nan, inplace=True)
            valid_df.dropna(inplace=True)
            # Resetting dataframe indexes for training anf validation dataframe
            train_df.reset_index(drop=True, inplace=True)
            valid_df.reset_index(drop=True, inplace=True)
            del temp_df
        if len(label_cols)>1:
            text_data.classes = label_cols
        else:
            text_data.classes = valid_df[label_cols[0]].unique().tolist()
        text_data._bs = batch_size
        text_data._text_cols = text_cols
        text_data._label_cols = label_cols
        text_data._train_df = train_df[[text_cols] + label_cols]
        text_data._valid_df = valid_df[[text_cols] + label_cols]
        text_data._training_indexes = range(0, train_df.shape[0])

        return text_data

    @classmethod
    def prepare_data_for_seq2seq(
            cls,
            data,
            text_cols,
            label_cols,
            train_file="train.csv",
            val_split_pct=0.1,
            seed=42,
            batch_size=8,
            process_labels=True,
            remove_html_tags=False,
            remove_urls=False,
        ):

        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        text_data = cls("sequence_translation")
        if not os.path.exists(data):
            raise Exception(f"Provided data directory - {data}, does not exists")

        training_file_path = os.path.join(data, train_file)

        if not os.path.exists(training_file_path):
            raise Exception(f"Provided data directory does not contain {train_file} file")

        train_df = read_file(training_file_path)
        train_df = cls._preprocess_df(train_df, text_cols, label_cols, process_labels, remove_html_tags, remove_urls)

        random.seed(seed)
        
        validation_indexes = random.sample(range(train_df.shape[0]), round(val_split_pct * train_df.shape[0]))
        training_indexes = list(set([i for i in range(train_df.shape[0])]) - set(validation_indexes))

        temp_df = copy.deepcopy(train_df)
        train_df = temp_df.loc[training_indexes]
        valid_df = temp_df.loc[validation_indexes]
        # Removing rows with empty strings in the text_cols from the training and validation data
        train_df[text_cols].replace('', np.nan, inplace=True)
        train_df.dropna(inplace=True)
        valid_df[text_cols].replace('', np.nan, inplace=True)
        valid_df.dropna(inplace=True)
        # Resetting dataframe indexes for training anf validation dataframe
        train_df.reset_index(drop=True, inplace=True)
        valid_df.reset_index(drop=True, inplace=True)
        del temp_df
        text_data._bs = batch_size
        text_data._text_cols = text_cols
        text_data._label_cols = label_cols
        text_data._train_df = train_df[[text_cols] + label_cols]
        text_data._valid_df = valid_df[[text_cols] + label_cols]
        text_data.val_split_pct = val_split_pct
        text_data._training_indexes = range(0, train_df.shape[0])
        text_data._model_type = None
        return text_data

    def get_databunch(self):
        if self._is_empty:
            return None
        else:
            return self._databunch

    def _prepare_databunch(self, tokenizer, vocab=None, pad_first=True, pad_idx=0, **kwargs):
        """
        Wrapper to create fastai TextDataBunch Object
        """
        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        logger = kwargs.get("logger")
        if logger:logger.info(f"Preparing databunch for task type - {self._task}")
        if self._task == "classification":
            self._databunch = TextClasDataBunch.\
                from_df(".",
                        train_df=self._train_df,
                        valid_df=self._valid_df,
                        tokenizer = tokenizer,
                        vocab=vocab,
                        include_bos=False,
                        include_eos=False,
                        text_cols=self._text_cols,
                        label_cols=self._label_cols,
                        bs=self._bs,
                        collate_fn=partial(pad_collate, pad_first=pad_first, pad_idx=pad_idx),
                        **self.databunch_kwargs
                        )

            TextList.show_xys = types.MethodType(show_text_xys, TextList)
            TextList.show_xyzs = types.MethodType(show_text_xyzs, TextList)
        elif self._task == "ner":
            model_type, seq_length = kwargs["model_type"], kwargs["seq_len"]
            dl_kwargs = {"pin_memory": self.databunch_kwargs.get('pin_memory')}
            device, num_workers = self.databunch_kwargs["device"], self.databunch_kwargs.get('num_workers')
            if num_workers: dl_kwargs["num_workers"] = num_workers

            train_ds = TransformerNERDataset(self._train_tokens, self._train_tags, tokenizer,
                                             self._label2id, model_type=model_type, seq_length=seq_length)
            valid_ds = TransformerNERDataset(self._valid_tokens, self._valid_tags, tokenizer,
                                             self._label2id, model_type=model_type, seq_length=seq_length)

            self._databunch = TransformerNERDataBunch.\
                create(train_ds, valid_ds, tokenizer, bs=self._bs, device=device, seq_len=seq_length, **dl_kwargs)
        else:
            raise Exception(f"Wrong task - {self._task} selected. Allowed values are 'ner', 'classification'")

        self._is_empty = False
        self._backbone = kwargs.get('backbone')

    def _prepare_seq2seq_databunch(self, transformer_processor, pad_first, pad_idx, **kwargs):
        """
        Wrapper to create fastai TextDataBunch Object
        """
        self._model_type = kwargs.get('model_type')
        
        if not HAS_FASTAI:
            return
        dl_tfms=None
        if self._model_type in ['t5', 'bart', 'marian']:
            dl_tfms=teacher_forcing_tfm
        data = SequenceToSequenceTextList.from_df(self._train_df,cols=self._text_cols, processor=transformer_processor)\
                        .split_by_rand_pct(valid_pct=self.val_split_pct)\
                        .label_from_df(cols= self._label_cols, label_cls=TextList, processor=transformer_processor)\
                        .databunch(bs=self._bs, pad_first=pad_first, pad_idx=pad_idx, dl_tfms=dl_tfms, **self.databunch_kwargs)
        
        self._is_empty = False
        self._databunch = data
        
        self._backbone = kwargs.get('backbone')

    @staticmethod
    def _preprocess_df(dataframe, text_cols, label_cols, process_labels, remove_html_tags=False, remove_urls=False):
        """
        Do some pre-processing on the dataframe columns
        """
        dataframe[text_cols] = dataframe.apply(
            lambda row: TextDataObject.process_text(row[text_cols], remove_html_tags, remove_urls), axis=1)
        if process_labels:
            for label in label_cols:
                dataframe[label] = dataframe.apply(
                    lambda row: TextDataObject.process_text(row[label], remove_html_tags, remove_urls), axis=1)

        return dataframe

    @staticmethod
    def process_text(text, remove_html_tags=False, remove_urls=False, remove_space_before_after_spl_char=False):
        """
        Perform some basic cleanup like removing HTML tags, removing urls,
        removing spaces before and after special characters
        converting multiple white spaces to single white space
        """
        if remove_urls: text = re.sub(r'\b(?:(?:https?|ftp)://)?\w[\w-]*(?:\.[\w-]+)+\S*', ' ', text)

        # text = re.sub(r'<.*?>', '', text)
        if remove_html_tags:
            if not HAS_BEAUTIFULSOUP:
                raise Exception("This module requires BeautifulSoup.")
            text = BeautifulSoup(text, 'html.parser').get_text(separator=" ", strip=True)

        if remove_space_before_after_spl_char:
            text = re.sub(' ([@.#$/:-]) ?', r'\1', text)

        if any([remove_html_tags, remove_urls, remove_space_before_after_spl_char]): text = re.sub(' +', ' ', text)
        return text.strip()

    def show_batch(self, rows=5, max_len=max_len):
        """
        Shows a batch of dataframe prepared without applying transforms.
        """
        if self._task == "classification":
            return self._classification_show_batch(rows=rows, max_len=max_len)
        elif self._task == "ner":
            return self._ner_show_batch(rows=rows)
        elif self._task == "sequence_translation":
            return self._classification_show_batch(rows=rows, max_len=max_len)
        else:
            raise Exception(f"Wrong task - {self._task} selected. Allowed values are 'ner', 'classification'")

    def _ner_show_batch(self, rows=5):
        results = []
        indexes = list(range(len(self._train_tokens)))
        random.shuffle(indexes)
        rows = min(len(indexes), rows)
        batch_tokens = [self._train_tokens[i] for i in indexes[:rows]]
        batch_labels = [self._train_tags[i] for i in indexes[:rows]]
        for index, (tokens, labels) in enumerate(zip(batch_tokens, batch_labels)):
            entity_dict = dict()
            entities = extract_entities(tokens, labels)
            text = process_text(" ".join(tokens))
            entity_dict["Text"] = text
            _ = [entity_dict.setdefault(x[1], []).append(process_text(" ".join(x[0]))) for x in entities if x[0]]
            results.append(entity_dict)
        df = pd.DataFrame(results, )
        df.fillna("", inplace=True)
        return df

    def _classification_show_batch(self, rows=5, max_len=max_len):
        processed_data = []
        rows = min(len(self._training_indexes), rows)
        random_batch = random.sample(self._training_indexes, rows)
        # rows = min(len(self._train_df), rows)
        # random_batch = np.random.randint(0,self._train_df.index.max(),rows)
        dataframe = self._train_df.loc[random_batch]
        if len(self._label_cols) > 1:
            for idx, item in dataframe.iterrows():
                target = ";".join([column for column in self._label_cols if int(getattr(item, column))])
                source = getattr(item, self._text_cols)
                if max_len is not None: source = " ".join(source.split(" "))[:max_len]
                processed_data.append([source, target])

            dataframe = pd.DataFrame(processed_data, columns=["source", "target"])

        return dataframe.style\
            .set_table_styles([dict(selector='th', props=[('text-align', 'left')])])\
            .set_properties(**{'text-align': "left"}).hide_index()

    def create_empty_object_for_ner(self, entities, address_tag, label2id, batch_size=4):
        self._bs = batch_size
        self._unique_tags = entities
        self._address_tag = address_tag
        self._label2id = label2id
        self._train_tags, self._train_tokens = [], []
        self._valid_tags, self._valid_tokens = [], []

    def create_empty_object_for_classification(self, text_cols, label_cols, classes, is_multilabel=False):
        self._text_cols = text_cols
        self._label_cols = label_cols

        text_list = TextList([], ignore_empty=True).split_by_idx([])
        if is_multilabel:
            self._databunch = text_list.label_const(0, label_cls=MultiCategoryList, classes=classes).databunch()
        else:
            self._databunch = text_list.label_const(0, label_cls=CategoryList, classes=classes).databunch()

        self._is_empty = False

    def create_empty_seq2seq_data(self, text_cols, label_cols):
        self._text_cols = text_cols
        self._label_cols = label_cols

        self._databunch = SequenceToSequenceTextList([])\
                        .split_none()\
                        .label_const()\
                        .databunch()
        self._is_empty = False

        # self._train_df = self._create_empty_df(batch_size, [text_cols] + label_cols)
        # self._valid_df = self._create_empty_df(batch_size, [text_cols] + label_cols)

    # @staticmethod
    # def _create_empty_df(batch_size, columns):
    #     l1 = ["" for x in columns]
    #     l2 = [l1 for x in range(batch_size)]
    #     return pd.DataFrame(l2, columns=columns)




import traceback
import warnings
warnings.filterwarnings("ignore", module='transformers')
from .._data import _raise_fastai_import_error
HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from .._utils.common import _get_device_id
    from fastprogress.fastprogress import progress_bar
    from transformers.modeling_auto import MODEL_FOR_QUESTION_ANSWERING_MAPPING
    EXPECTED_MODEL_TYPES = [x.__name__.replace('Config', '') for x in MODEL_FOR_QUESTION_ANSWERING_MAPPING.keys()]
except Exception as e:
    transformer_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []


class QuestionAnswering:
    """
    Creates a `QuestionAnswering` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify the HuggingFace
                            transformer model name which will be used to
                            extract the answers from a given passage/context.

                            To learn more about the available models for
                            question-answering task, kindly visit:-
                            https://huggingface.co/models?filter=question-answering
    =====================   ===========================================

    :returns: `QuestionAnswering` Object
    """

    #: supported transformer backbones
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)

        logger = logging.get_logger()
        logger.setLevel(logging.ERROR)
        self._device = _get_device_id()
        self._task = "question-answering"
        try:
            self.model = pipeline(self._task, model=backbone, device=self._device)
        except Exception as e:
            error_message = (f"Model - `{backbone}` cannot be used for {self._task} task.\n"
                             f"Model type should be one of {EXPECTED_MODEL_TYPES}.")
            raise Exception(error_message)

    def get_answer(self, text_or_list, context, **kwargs):
        """
        Find answers for the asked questions from the given passage/context

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. Questions or a list
                                of questions one wishes to seek an answer for.
        ---------------------   -------------------------------------------
        context                 Required string. The context associated with
                                the question(s) which contains the answers.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        num_answers             Optional integer. The number of answers to
                                return. The answers will be chosen by order
                                of likelihood.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        max_answer_length       Optional integer. The maximum length of the
                                predicted answers.
                                Default value is set to 15.
        ---------------------   -------------------------------------------
        max_question_length     Optional integer. The maximum length of the
                                question after tokenization. Questions will be
                                truncated if needed.
                                Default value is set to 64.
        ---------------------   -------------------------------------------
        impossible_answer       Optional bool. Whether or not we accept impossible
                                as an answer.
                                Default value is set to False
        =====================   ===========================================

        :returns: a list or a list of list containing the answer(s) for the input question(s)
        """
        results, kwargs_dict = [], {}

        kwargs_dict["topk"] = kwargs.get("num_answers", 1)
        kwargs_dict["max_answer_len"] = kwargs.get("max_answer_length", 15)
        kwargs_dict["max_question_len"] = kwargs.get("max_question_length", 64)
        kwargs_dict["handle_impossible_answer"] = kwargs.get("impossible_answer", False)

        if not isinstance(text_or_list, (list, tuple)): text_or_list = [text_or_list]
        for i in progress_bar(range(len(text_or_list))):
            results.append(self.model(question=text_or_list[i], context=context, **kwargs_dict))

        return self._process_result(results, text_or_list)

    @staticmethod
    def _process_result(result_list, question_list):
        processed_results = []
        for result, question in zip(result_list, question_list):
            if isinstance(result, dict):
                tmp_dict = {"question": question, "answer": result["answer"], "score": result["score"]}
                processed_results.append(tmp_dict)
            elif isinstance(result, list):
                item_list = [{"question": question, "answer": item["answer"], "score": item["score"]} for item in result]
                processed_results.append(item_list)

        return processed_results



import json
import logging
import random
import traceback
from functools import partial
from pathlib import Path

from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel
from ._codetemplate import image_classifier_prf

HAS_OPENCV = True
HAS_FASTAI = True
HAS_ARCPY = True

# Try to import the necessary modules
# Exception will turn the HAS_FASTAI flag to false
# so that relevant exception can be raised
try:
    import numpy as np
    from fastai.vision.learner import create_body
    from torch import optim
    from torchvision import models
    from fastai.basic_train import Learner
    from arcgis.learn.models._linknet_utils import (
        LinkNetModel,
        accuracy,
        miou,
    )
    from fastai.basic_data import DatasetType
    from arcgis.learn._utils.segmentation_loss_functions import mIoULoss
    from arcgis.learn.models._arcgis_model import _resnet_family, _EmptyData
    from arcgis.learn._utils.common import (
        get_multispectral_data_params_from_emd,
        safe_json,
    )
    from fastai.callback import annealing_no
    from fastai.callbacks.general_sched import TrainingPhase, GeneralScheduler
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class LinkNet(ArcGISModel):
    """
    Model architecture from https://arxiv.org/pdf/1707.03718.pdf.
    Creates a LinkNet Image Segmentation / Pixel Classification model.
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `LinkNet`, which
                            is `resnet34` by default.
                            Compatible backbones: 'resnet18', 'resnet34',
                            'resnet50', 'resnet101', 'resnet152'
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================
    :returns: `LinkNet` Object
    """

    def __init__(
        self, data, backbone=None, pretrained_path=None, *args, **kwargs,
    ):
        # Set default backbone to be 'resnet34'
        if backbone is None:
            backbone = models.resnet34
        super().__init__(data, backbone, **kwargs)

        self._ignore_classes = kwargs.get("ignore_classes", [])
        if self._ignore_classes != [] and len(self._data.classes) <= 3:
            raise Exception(
                "`ignore_classes` parameter can only be used when the dataset has more than 2 classes."  # noqa
            )

        data_classes = list(self._data.class_mapping.keys())
        if 0 not in list(data.class_mapping.values()):
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes if k != 0
            ]
        else:
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes
            ]
        if self._ignore_classes != []:
            if 0 not in self._ignore_mapped_class:
                self._ignore_mapped_class.insert(0, 0)

        pixel_accuracy = partial(
            accuracy, ignore_mapped_class=self._ignore_mapped_class
        )
        mean_iou = partial(miou, ignore_mapped_class=self._ignore_mapped_class)

        n_bands = len(getattr(self._data, "_extract_bands", [0, 1, 2]))
        _backbone = self._backbone
        if hasattr(self, "_orig_backbone"):
            _backbone = self._orig_backbone

        # Check if a backbone provided is compatible, use resnet50 as default
        if not self._check_backbone_support(_backbone):
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"  # noqa
            )

        self.name = "LinkNet"
        self._code = image_classifier_prf

        self._chip_size = (self._data.chip_size, self._data.chip_size)

        # Cut-off the backbone before the penultimate layer
        self._encoder = create_body(self._backbone, -2)

        # Initialize the model, loss function and the Learner object
        self._model = LinkNetModel(
            self._encoder, n_classes=self._data.c,
            chip_size=self._chip_size, n_bands=n_bands,
        )
        self._loss_f = mIoULoss(n_classes=self._data.c)
        learner_kwargs = {}
        self._opt_func_name = kwargs.get("opt_func", None)
        if self._opt_func_name:
            _opt_func = self._get_optimizer(self._opt_func_name)
            if _opt_func:
                self._opt_func_args = kwargs.get("opt_func_args", {})
                learner_kwargs["opt_func"] = partial(_opt_func, **self._opt_func_args)

        self.learn = Learner(
            self._data, self._model, loss_func=self._loss_f,
            metrics=[pixel_accuracy, mean_iou], **learner_kwargs,
        )
        if pretrained_path is not None:
            self.load(str(pretrained_path))
        self._arcgis_init_callback()  # make first conv weights learnable

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    def _get_optimizer(self, optim_name):
        optim_funcs = {"sgd": optim.SGD}
        return optim_funcs.get(optim_name, None)

    # Return a list of supported backbones names
    @property
    def supported_backbones(self):
        """
        Supported torchvision backbones for this model.
        """
        return LinkNet._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    def _get_emd_params(self):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        _emd_template["ModelConfiguration"] = "_LinkNet_Inference"
        _emd_template["ExtractBands"] = [0, 1, 2]
        model_params = {
            "backbone": self._backbone,
            "backend": self._backend,
            "opt_func": self._opt_func_name if hasattr(self, "_opt_func_name") else "",
            "opt_func_args": self._opt_func_args
            if hasattr(self, "_opt_func_args")
            else "",
        }
        _emd_template["ModelParameters"] = model_params
        _emd_template["Classes"] = []

        class_data = {}
        # 0th index is background
        for _, class_name in enumerate(self._data.classes[1:]):
            inverse_class_mapping = {
                v: k for k, v in self._data.class_mapping.items()
            }
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        if hasattr(self._data, "_base"):
            data_params = {}
            for param_key, param_value in self._data._base.__dict__.items():
                if "files" in param_key:
                    continue
                if hasattr(param_value, "__dict__"):
                    continue
                if not safe_json(param_value):
                    param_value = str(param_value)
                data_params[f"{param_key}"] = param_value
            _emd_template[self._data._base.__class__.__name__] = data_params

        return _emd_template

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a LinkNet Pixel Classifier from an Esri Model Definition (EMD) file.
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        :returns: `LinkNet` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = Path(emd_path)
        emd = json.load(open(emd_path))
        model_file = Path(emd["ModelFile"])
        chip_size = emd["ImageWidth"]
        model_params = emd["ModelParameters"]

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=chip_size,
            )
            data.class_mapping = class_mapping
            data = get_multispectral_data_params_from_emd(data, emd)
            data.emd_path = emd_path
            data.emd = emd

        return cls(data, **model_params, pretrained_path=str(model_file))

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        ds_type = kwargs.get("ds_type", DatasetType.Valid)
        min_data_rows = min(len(self._data.dl(ds_type)), self._data.dl(ds_type).batch_size)
        if rows > min_data_rows:
            rows = min_data_rows
        self.learn.show_results(rows=rows, **kwargs)

    @property
    def _model_metrics(self):
        acc, mean_iou = self._get_model_metrics()
        return {
            "accuracy": "{0:1.4e}".format(acc),
            "mIoU": "{0:1.4e}".format(mean_iou),
        }

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get("checkpoint", True)
        if not hasattr(self.learn, "recorder"):
            return 0.0

        try:
            model_accuracy = self.learn.recorder.metrics[-1][0]
            model_iou = self.learn.recorder.metrics[-1][1]
            if checkpoint:
                metrics_array = np.array(self.learn.recorder.metrics)
                model_iou = np.max(metrics_array[:, 1])
                model_accuracy = metrics_array[np.argmax(metrics_array[:, 1]), 0]
        except BaseException:
            logger = logging.getLogger()
            logger.debug("Cannot retrieve model accuracy.")
            model_accuracy = 0.0

        return float(model_accuracy), float(model_iou)



import uuid


class DatePicker(object):
    """
    Creates a Date Selector widget for Side Panel or Header.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    range                       Optional boolean. True to create a range
                                selector.
    -------------------------   -------------------------------------------
    operator                    Optional String. Operator for non range
                                datepicker.
                                Options: "is", "is not", "is before",
                                "is or is before", "is after",
                                "is or is after".
    -------------------------   -------------------------------------------
    label                       Optional String. Label for the widget.
    -------------------------   -------------------------------------------
    **kwargs                    If "range" is True, provide two parameters
                                "min_value" and "max_value". If "range" is
                                False provide single parameter "value".
                                Allowed values:
                                None, "Today", or a fixed value in
                                24 hours format
                                (year, month, day, hour, minutes)
                                or
                                (year, month, day)
    =========================   ===========================================
    """
    def __init__(self, range=False, operator='is', label="", **kwargs):
        self._id = str(uuid.uuid4())
        self.type = "dateSelectorWidget"

        self._operator_mapping = {
            'is': 'is_on',
            'is not': 'is_not_on',
            'is before': 'is_before',
            'is or is before': 'is_on_before',
            'is after': 'is_after',
            'is or is after': 'is_on_after'
        }

        self._selection_type = "single"
        if range:
            self._selection_type = "range"
            self.operator = "between"
            min_value = kwargs.get("min_value", None)
            max_value = kwargs.get("max_value", None)
            if min_value:
                if isinstance(min_value, str) and min_value.lower() == "today":
                    self._min_value = {"type": "date", "includeTime": False, "defaultToToday": True}
                else:
                    self._min_value = {"type": "date", "includeTime": False, "defaultToToday": False, "year": min_value[0], "month": min_value[1]-1, "date": min_value[2], "hours": min_value[3] if len(min_value) > 3 else 0, "minutes": min_value[4] if len(min_value) > 3 else 0, "seconds": 0, "milliSeconds": 0}
            else:
                #raise Exception("Please provide min_value parameter")
                self._min_value = None

            if max_value:
                if isinstance(max_value, str) and max_value.lower() == "today":
                    self._max_value = {"type": "date", "includeTime": False, "defaultToToday": True}
                else:
                    self._max_value = {"type": "date", "includeTime": False, "defaultToToday": False, "year": max_value[0], "month": max_value[1]-1, "date": max_value[2], "hours": max_value[3] if len(max_value) > 3 else 0, "minutes": max_value[4] if len(max_value) > 3 else 0, "seconds": 0, "milliSeconds": 0}
            else:
                #raise Exception("Please provide max_value parameter")
                self._max_value = None
        else:
            self.operator = operator
            min_value = kwargs.get("value", None)
            self._max_value = None
            if min_value:
                if isinstance(min_value, str) and min_value.lower() == "today":
                    self._min_value = {"type": "date", "includeTime": False, "defaultToToday": True}
                else:
                    self._min_value = {"type": "date", "includeTime": False, "defaultToToday": False, "year": min_value[0], "month": min_value[1]-1, "date": min_value[2], "hours": min_value[3] if len(min_value) > 3 else 0, "minutes": min_value[4] if len(min_value) > 3 else 0, "seconds": 0, "milliSeconds": 0}
            else:
                self._min_value = None
                #raise Exception("Please provide value parameter")

        self.label = label

    def _convert_to_json(self):

        if self._selection_type == "range":
            self._operator_logic = "between"
        else:
            self._operator_logic = self._operator_mapping.get(self.operator, 'is_on')

        data = {
            "type": "dateSelectorWidget",
            "optionType": "datePicker",
            "datePickerOption": {
                "type": "datePicker",
                "selectionType": self._selection_type,
                "operator": self._operator_logic
            },
            "id": self._id,
            "name": "Date Selector (1)",
            "caption": self.label,
            "showLastUpdate": True,
            "noDataVerticalAlignment": "middle",
            "showCaptionWhenNoData": True,
            "showDescriptionWhenNoData": True
        }

        if self._min_value:
            data['datePickerOption']['minDefaultValue'] = self._min_value

        if self._max_value:
            data['datePickerOption']['maxDefaultValue'] = self._max_value

        return data

    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        from arcgis.apps.dashboard import SidePanel
        sp = SidePanel()
        sp.add_selector(self)
        url = Dashboard._publish_random(sp)
        return f"""<iframe src={url} width=300 height=300>"""


class NumberSelector(object):
    """
    Creates a Number Selector widget for Side Panel or Header.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    range                       Optional boolean. True to create a range
                                selector.
    -------------------------   -------------------------------------------
    display_type                Optional String. Display type can be from
                                "spinner", "slider", "input".
    -------------------------   -------------------------------------------
    label                       Optional string. Label for the selector.
    =========================   ===========================================

    **Keyword Arguments**

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    operator                    Optional string for non-range input.
                                Allowed: "equal", "not equal", "greater than",
                                "greater than or equal", "less than",
                                "less than or equal".
                                Default: "equal"
    -------------------------   -------------------------------------------
    increment_factor            Optional int for slider and spinner input.
    =========================   ===========================================
    """

    def __init__(self, range=False, display_type="spinner", label='Select a number', **kwargs):

        self._json = {}
        self._display_type = display_type
        self.type = "numberSelectorWidget"
        self._id = str(uuid.uuid4())
        self._label = label if label else ''
        self._increment = kwargs.get('increment_factor', 1)
        self._lower_limit = 0
        self._upper_limit = 100
        self._lower_default = 0
        self._upper_default = 100
        self._range = range
        self._operator = kwargs.get('operator', "equal")

        self._left_placeholder_text = ""
        self._right_placeholder_text = ""

        self._dataset = None
        self._operator_mapping = {
            "equal": "equal",
            "not equal": "not_equal",
            "greater than": "greater",
            "greater than or equal": "greater_or_equal",
            "less than": "less",
            "less than or equal": "less_or_equal"
        }

        self._constraint = {
            "type": "fixed",
            "lowerLimit": self._lower_limit,
            "upperLimit": self._upper_limit,
            "firstDefault": self._lower_default,
            "secondDefault": self._upper_default
        }

    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        from arcgis.apps.dashboard import SidePanel
        sp = SidePanel()
        sp.add_selector(self)
        url = Dashboard._publish_random(sp)
        return f"""<iframe src={url} width=300 height=300>"""

    @property
    def placeholder_text(self):
        """
        :return: Text for left place holder in range type or default place holder.
        """
        return self._left_placeholder_text

    @placeholder_text.setter
    def placeholder_text(self, value):
        """
        Text for left place holder in range type or default place holder.
        """
        self._left_placeholder_text = value

    @property
    def right_placeholder_text(self):
        """
        :return: Text for right place holder in range type.
        """
        return self._right_placeholder_text

    @right_placeholder_text.setter
    def right_placeholder_text(self, value):
        """
        Text for right place holder in range type.
        """
        self._right_placeholder_text = value

    def set_statistics_limits(self, item, field, default="min", layer_id=0):
        """
        Set the item to pick values from for spinner and slider display type.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        item                        Required Portal Item. Item to pick values from.
        -------------------------   -------------------------------------------
        field                       Required String. Field from the Portal Item.
        -------------------------   -------------------------------------------
        default                     Optional String. Default value statistic.
                                    Options: "min", "max", "avg"
        -------------------------   -------------------------------------------
        layer_id                    Optional Int. Layer Id for the item.
        =========================   ===========================================
        """

        self._stat_dataset = {
            "type": "serviceDataset",
            "dataSource": {
                "type": "featureServiceDataSource",
                "itemId": item.itemid,
                "layerId": layer_id,
                "table": True
            },
            "outFields": ["*"],
            "groupByFields": [],
            "orderByFields": [],
            "statisticDefinitions": [
                {"onStatisticField": field, "outStatisticFieldName": "lowerLimit",
                 "statisticType": "min"},
                {"onStatisticField": field, "outStatisticFieldName": "upperLimit",
                 "statisticType": "max"},
                {"onStatisticField": field, "outStatisticFieldName": "averageStatisticValue",
                 "statisticType": "avg"}
            ],
            "maxFeatures": 50,
            "querySpatialRelationship": "esriSpatialRelIntersects",
            "returnGeometry": False,
            "clientSideStatistics": False,
            "name": "main"
        }

        self._constraint = {"type": "statistic", "defaultStatistic": "min"}

        if default in ["min", "max", "avg"]:
            self._constraint["defaultStatistic"] = default
    
    def set_defined_limits(self, lower_limit=0, upper_limit=100, **kwargs):
        """
        Set the item to pick values from for spinner and slider display type.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        lower_limit                 Optional integer. Set the lower limit.
        -------------------------   -------------------------------------------
        upper_limit                 Optional integer. Set the upper limit.
        =========================   ===========================================

        **Keyword Arguments**

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        default                     Optional integer. Set default value for
                                    non-range selector.
        -------------------------   -------------------------------------------
        lower_default               Optional integer. Set the lower default
                                    value for range selector.
        -------------------------   -------------------------------------------
        upper_default               Optional integer. Set the upper default
                                    value for range selector.
        =========================   ===========================================
        """

        self._lower_limit = lower_limit
        self._upper_limit = upper_limit
        self._lower_default = kwargs.get('default', kwargs.get('lower_default', 0))
        self._upper_default = kwargs.get('upper_default', 100)
        self._constraint = {
            "type": "fixed",
            "lowerLimit": self._lower_limit,
            "upperLimit": self._upper_limit,
            "firstDefault": self._lower_default,
            "secondDefault": self._upper_default
        }

    def _convert_to_json(self):

        # self._dataset = {
        #     "type": "serviceDataset",
        #     "dataSource": {
        #         "type": "featureServiceDataSource",
        #         "itemId": self.item.itemid,
        #         "layerId": 0,
        #         "table": True
        #     },
        #     "outFields": ["*"],
        #     "groupByFields": [],
        #     "orderByFields": [],
        #     "statisticDefinitions": [
        #         {"onStatisticField": field, "outStatisticFieldName": "lowerLimit",
        #          "statisticType": "min"},
        #         {"onStatisticField": field, "outStatisticFieldName": "upperLimit",
        #          "statisticType": "max"},
        #         {"onStatisticField": field, "outStatisticFieldName": "averageStatisticValue",
        #          "statisticType": "avg"}
        #     ],
        #     "maxFeatures": 50,
        #     "querySpatialRelationship": "esriSpatialRelIntersects",
        #     "returnGeometry": False,
        #     "clientSideStatistics": False,
        #     "name": "main"
        # }

        json = {
            "type": "numericSelectorWidget",
            "displayType": self._display_type,
            "increment": self._increment,
            "valueLabelFormat": {"name": "value", "type": "decimal", "prefix": False, "pattern": "#,###"},
            "selection": {"type": "single" if not self._range else "range"},
            "datasets": [],
            "id": self._id,
            "name": "Number Selector (1)",
            "caption": self._label,
            "showLastUpdate": True,
            "noDataVerticalAlignment": "middle",
            "showCaptionWhenNoData": True,
            "showDescriptionWhenNoData": True
        }

        if not self._range:
            json['selection']['operator'] = self._operator

        if self._constraint:
            json["constraint"] = self._constraint

        if self._display_type == "input":
            json['selection']['placeholderText'] = self._left_placeholder_text
            if self._range:
                json['selection']['rightPlaceHolderText'] = self._right_placeholder_text

        if getattr(self,'_stat_dataset', None):
            json["datasets"].append(self._stat_dataset)
        # else:
        #     json["datasets"].append(self._dataset)

        return json


class CategorySelector(object):
    """
    Creates a Category Selector widget for Side Panel or Header.
    """

    def __init__(self):
        self._categories_from = 'static'

        self._id = str(uuid.uuid4())
        self._selector = CategorySelectorProperties._create_selector_properties()

        self._dataset = None

    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        from arcgis.apps.dashboard import SidePanel
        sp = SidePanel()
        sp.add_selector(self)
        url = Dashboard._publish_random(sp)
        return f"""<iframe src={url} width=300 height=300>"""

    def set_defined_values(self, key_value_pairs, value_type="string"):
        """
        Set defined values for the dropdown.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        key_value_pairs             Optional list of tuples. The tuple should
                                    contain labels and their corresponding values.
        -------------------------   -------------------------------------------
        value_type                  Optional String.
                                    The data type of the values in the tuple.
                                    "integer" or "string
        =========================   ===========================================
        """
        type_caster = str
        self._categories_from = "static"
        if value_type == "integer":
            type_caster = int

        if value_type not in ["string", "integer"]:
            value_type = "string"

        self._dataset = {
            "type":"staticDataset",
            "data": {
                "type": "staticValues",
                "dataType": value_type,
                "values": []
            },
            "name": "main"
        }

        id = 0
        for pair in key_value_pairs:
            self._dataset['data']['values'].append({
                "type":" labelledValue",
                "id": str(id),
                "label": pair[0],
                "value": type_caster(pair[1])
            })
            id = id + 1

    def set_feature_options(self, item, line_item_text="", field_name=None, max_features=50):
        """
        Set feature values for dropdown.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        item                        Required Portal Item.
                                    Dropdown values will be populated from this.
        -------------------------   -------------------------------------------
        line_item_text              Optional String.
                                    This text will be displayed with options.
        -------------------------   -------------------------------------------
        field_name                  Optional String.
                                    Data from this field will be added to list.
        -------------------------   -------------------------------------------
        max_features                Optional Integer.
                                    Set max features to display.
        =========================   ===========================================
        """
        self._categories_from = "features"

        self._line_item_text = line_item_text if line_item_text else ""
        if field_name is not None:
            self._line_item_text = self._line_item_text + '{' + field_name + '}'

        self._dataset = {
            "type": "serviceDataset",
            "dataSource": {
                "type": "featureServiceDataSource",
                "itemId": item.itemid,
                "layerId": 0,
                "table": True
            },
            "outFields": ["*"],
            "groupByFields": [],
            "orderByFields": [],
            "statisticDefinitions": [],
            "maxFeatures": max_features,
            "querySpatialRelationship": "esriSpatialRelIntersects",
            "returnGeometry": False,
            "clientSideStatistics": False,
            "name": "main"
        }

    def set_group_by_values(self, item, category_field, max_features=50):
        """
        Set group by values for dropdown.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        item                        Required Portal Item.
                                    Dropdown values will be populated from this.
        -------------------------   -------------------------------------------
        category_field              Optional String. This string denotes the
                                    field to pick the values from.
        -------------------------   -------------------------------------------
        max_features                Optional Integer.
                                    Set max features to display.
        =========================   ===========================================
        """
        self._categories_from = "groupByValues"

        self._dataset = {
            "type": "serviceDataset",
            "dataSource": {
                "type": "featureServiceDataSource",
                "itemId": item.itemid,
                "layerId": 0,
                "table": True
            },
            "outFields": ["*"],
            "groupByFields": [category_field],
            "orderByFields": [category_field + " asc"],
            "statisticDefinitions": [
                {
                    "onStatisticField": category_field,
                    "outStatisticFieldName": "count_result",
                    "statisticType": "count"
                }
            ],
            "maxFeatures": max_features,
            "querySpatialRelationship": "esriSpatialRelIntersects",
            "returnGeometry": False,
            "clientSideStatistics": False,
            "name": "main"
        }

    @property
    def selector(self):
        """
        :return: Selector Properties Object, set label, preferred display, display threshold, operator etc.
        """
        return self._selector

    def _convert_to_json(self):
        json = {
            "type": "categorySelectorWidget",
            "category": {},
            "selection": {"type": self._selector._selection_type, "defaultSelection": "0", "operator": self._selector._operator},
            "preferredDisplayType": self._selector._preferred_display, #dropdown, button_bar, radio_buttons
            "displayThreshold": self._selector._display_threshold,
            "datasets": [],
            "id": self._id,
            "name": "Category Selector (1)",
            "caption": self._selector._label,
            "showLastUpdate": True,
            "noDataVerticalAlignment": "middle",
            "showCaptionWhenNoData": True,
            "showDescriptionWhenNoData": True
        }

        if self._selector._none_option:
            json["noneLabelPlacement"] = self._selector._none_placement
            json["noneLabel"] = self._selector._none_label

        if self._categories_from == "static":
            json["category"] = {"type": "static"}
        elif self._categories_from == "features":
            json["category"] = {"type": "features", "itemText": self._line_item_text}
        elif self._categories_from == "groupByValues":
            json["category"] = {"type": "groupByValues", "nullLabel": "Null", "blankLabel": "Blank", "labelOverrides": []}

        if self._dataset:
            json["datasets"].append(self._dataset)

        return json


class CategorySelectorProperties(object):

    @classmethod
    def _create_selector_properties(cls):
        selector = cls()
        selector._preferred_display = "dropdown"
        selector._default_selection = 0
        selector._label = ""
        selector._selection_type = "single"
        selector._operator = "equal"
        selector._display_threshold = 10

        selector._none_option = False
        selector._none_placement = "first"
        selector._none_label = "None"

        return selector

    @property
    def preferred_display(self):
        """
        :return: Preferred display for the selector.
        """
        return self._preferred_display

    @preferred_display.setter
    def preferred_display(self, value):
        """
        Set preferred display from "dropdown", "button_bar" or "radio_buttons"
        """
        if value not in ["dropdown", "button_bar", "radio_buttons"]:
            raise Exception("Invalid preferred display")

        self._preferred_display = value

    @property
    def label(self):
        """
        :return: Label Text
        """
        return self._label

    @label.setter
    def label(self, value):
        """
        Set label text.
        """
        self._label = value

    @property
    def multiple_selection(self):
        """
        :return: True or False for multiple selection
        """
        if self._selection_type == "single":
            return False

        return True

    @multiple_selection.setter
    def multiple_selection(self, value):
        """
        Set selection type to True or False.
        """
        if value:
            self._selection_type = "multiple"
            if self._operator == "equal":
                self._operator = "is_in"
            elif self._operator == "not_equal":
                self._operator = "is_not_in"
        else:
            self._selection_type = "single"
            if self._operator == "is_in":
                self._operator = "equal"
            elif self._operator == "is_not_in":
                self._operator = "not_equal"

    @property
    def include_values(self):
        """
        :return: True if values selected are to be taken.
        """
        if self._operator in ["is_in", "equal"]:
            return True

        return False

    @include_values.setter
    def include_values(self, value):
        """
        Set True to take selected values, False to take unselected values.
        """
        if value:
            if self._selection_type == "multiple":
                self._operator = "is_in"
            else:
                self._operator = "equal"
        else:
            if self._selection_type == "multiple":
                self._operator = "is_not_in"
            else:
                self._operator = "not_equal"

    @property
    def display_threshold(self):
        """
        :return: Return display threshold.
        """
        return self._display_threshold

    @display_threshold.setter
    def display_threshold(self, value):
        """
        Set Dropdown display threshold
        """
        self._display_threshold = value

    @property
    def none(self):
        """
        :return: Label for None option if set else None.
        """
        return self._none_label if self._none_option == True else None

    @none.setter
    def none(self, value):
        """
        Set Label for None option. Set None to disable
        """
        if value is None:
            self._none_option = False
            self._none_label = ""
        else:
            self._none_option = True
            self._none_label = value

    @property
    def none_placement(self):
        """
        :return: None Placement "first" or "last".
        """
        return self._none_placement

    @none_placement.setter
    def none_placement(self, value):
        """
        Set none option placement to "first" or "last".
        """

        if value not in ["first", "last"]:
            raise Exception("Invalid value")

        self._none_placement = value

